# pylint: disable=too-many-lines
import json
from array import array
from collections import Counter, defaultdict
from functools import reduce
from itertools import chain, filterfalse, product
from operator import mul, xor
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
)

import sympy
from typing_extensions import TypedDict

from comb_spec_searcher import CombinatorialClass, VerificationStrategy
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.typing import Parameters
from permuta import Perm
from permuta.misc import DIR_EAST, DIR_WEST

from .algorithms import (
    AllObstructionInferral,
    ComponentFusion,
    EmptyCellInferral,
    Factor,
    FactorWithInterleaving,
    FactorWithMonotoneInterleaving,
    Fusion,
    GriddedPermReduction,
    GriddedPermsOnTiling,
    MinimalGriddedPerms,
    ObstructionTransitivity,
    RequirementPlacement,
    RowColMap,
    RowColSeparation,
    SubclassVerificationAlgorithm,
    SubobstructionInferral,
    guess_obstructions,
)
from .assumptions import (
    ComponentAssumption,
    SkewComponentAssumption,
    SumComponentAssumption,
    TrackingAssumption,
)
from .exception import InvalidOperationError
from .griddedperm import GriddedPerm
from .gui_launcher import run_gui
from .misc import intersection_reduce, union_reduce

__all__ = ["Tiling"]


Cell = Tuple[int, int]
ReqList = Tuple[GriddedPerm, ...]
CellBasis = Dict[Cell, Tuple[List[Perm], List[Perm]]]
CellFrozenSet = FrozenSet[Cell]
Dimension = Tuple[int, int]


CachedProperties = TypedDict(
    "CachedProperties",
    {
        "active_cells": CellFrozenSet,
        "backward_map": RowColMap,
        "cell_basis": CellBasis,
        "dimensions": Dimension,
        "empty_cells": CellFrozenSet,
        "forward_map": RowColMap,
        "point_cells": CellFrozenSet,
        "positive_cells": CellFrozenSet,
        "possibly_empty": CellFrozenSet,
    },
    total=False,
)


class Tiling(CombinatorialClass):
    """Tiling class.

    Zero-indexed coordinates/cells from bottom left corner where the (x, y)
    cell is the cell in the x-th column and y-th row.

    Tilings store the obstructions and requirements but also caches the empty
    cells and the active cells.
    """

    def __init__(
        self,
        obstructions: Iterable[GriddedPerm] = tuple(),
        requirements: Iterable[Iterable[GriddedPerm]] = tuple(),
        assumptions: Iterable[TrackingAssumption] = tuple(),
        remove_empty_rows_and_cols: bool = True,
        derive_empty: bool = True,
        simplify: bool = True,
        sorted_input: bool = False,
        already_minimized_obs: bool = False,
    ) -> None:
        """
        - if remove_empty_rows_and_cols, then remove empty rows and columns.
        - if derive empty, then assume non-active cells are empty
        - if simplify, then obstructions and requirements will be simplifed
        - if not sorted_input, input will be sorted
        - already_minimized_obs indicates if the obstructions are already minimized
            we pass this through to GriddedPermReduction
        """
        self._cached_properties: CachedProperties = {}

        super().__init__()
        if sorted_input:
            # Set of obstructions
            self._obstructions = tuple(obstructions)
            # Set of requirement lists
            self._requirements = tuple(tuple(r) for r in requirements)
            # Set of assumptions
            self._assumptions = tuple(assumptions)
        else:
            # Set of obstructions
            self._obstructions = tuple(sorted(obstructions))
            # Set of requirement lists
            self._requirements = Tiling.sort_requirements(requirements)
            # Set of assumptions
            self._assumptions = tuple(sorted(assumptions))

        # Simplify the set of obstructions and the set of requirement lists
        if simplify:
            self._simplify_griddedperms(already_minimized_obs=already_minimized_obs)

        if not any(ob.is_empty() for ob in self.obstructions):
            # Remove gridded perms that avoid obstructions from assumptions
            if simplify:
                self.clean_assumptions()

            # Fill empty
            if derive_empty:
                if "empty_cells" not in self._cached_properties:
                    self._prepare_properties()

            # Remove empty rows and empty columns
            if remove_empty_rows_and_cols:
                self._remove_empty_rows_and_cols()

        else:
            self._obstructions = (GriddedPerm.empty_perm(),)
            self._requirements = tuple()
            self._assumptions = tuple()
            self._cached_properties["active_cells"] = frozenset()
            self._cached_properties["backward_map"] = RowColMap.identity((0, 0))
            self._cached_properties["cell_basis"] = {(0, 0): ([Perm()], [])}
            self._cached_properties["dimensions"] = (1, 1)
            self._cached_properties["empty_cells"] = frozenset([(0, 0)])
            self._cached_properties["forward_map"] = RowColMap.identity((0, 0))
            self._cached_properties["point_cells"] = frozenset()
            self._cached_properties["positive_cells"] = frozenset()
            self._cached_properties["possibly_empty"] = frozenset()

    @classmethod
    def from_perms(
        cls,
        obstructions: Iterable[Perm] = tuple(),
        requirements: Iterable[Iterable[Perm]] = tuple(),
    ) -> "Tiling":
        """
        Return a 1x1 tiling from that avoids permutation in `obstructions`
        and contains one permutation form each iterable of `requirements`.
        """
        t = Tiling(
            obstructions=(GriddedPerm(p, ((0, 0),) * len(p)) for p in obstructions)
        )
        for perms in requirements:
            req_list = [GriddedPerm(p, ((0, 0),) * len(p)) for p in perms]
            t = t.add_list_requirement(req_list)
        return t

    def _prepare_properties(self) -> None:
        """
        Compute _active_cells, _empty_cells, _dimensions, and store them
        """
        active_cells = union_reduce(
            set(ob.pos) for ob in self.obstructions if len(ob) > 1
        )
        active_cells.update(
            *(union_reduce(set(comp.pos) for comp in req) for req in self.requirements)
        )

        max_row = 0
        max_col = 0
        for cell in active_cells:
            max_col = max(max_col, cell[0])
            max_row = max(max_row, cell[1])
        dimensions = (max_col + 1, max_row + 1)

        empty_cells = tuple(
            cell
            for cell in product(range(dimensions[0]), range(dimensions[1]))
            if cell not in active_cells
        )

        # If the first obstruction is the empty perm, we shouldn't do any of this.
        if len(self._obstructions) > 0 and len(self._obstructions[0]) > 0:
            # We can assume that self._obstructions is sorted at this point, so to
            #   extract the point obstructions, we just pass though them until we've
            #   found the last one, then we slice the list there.
            index = 0
            for ob in self._obstructions:
                if len(ob) > 1:
                    break
                index += 1  # Now the last point obstruction is at index [index-1]
            non_point_obstructions = self._obstructions[index:]

            new_point_obstructions = tuple(
                GriddedPerm((0,), (cell,)) for cell in empty_cells
            )
            self._obstructions = new_point_obstructions + non_point_obstructions

        self._cached_properties["active_cells"] = frozenset(active_cells)
        self._cached_properties["empty_cells"] = frozenset(empty_cells)
        self._cached_properties["dimensions"] = dimensions

    def _simplify_griddedperms(self, already_minimized_obs=False) -> None:
        """
        Simplifies the set of obstructions and the set of requirement lists.
        The set of obstructions are first reduced to a minimal set. The
        requirements that contain any obstructions are removed from their
        respective lists. If any requirement list is empty, then the tiling is
        empty.
        """
        GPR = GriddedPermReduction(
            self.obstructions,
            self.requirements,
            sorted_input=True,
            already_minimized_obs=already_minimized_obs,
        )

        self._obstructions = GPR.obstructions
        self._requirements = GPR.requirements

    def _remove_empty_rows_and_cols(self) -> None:
        """Remove empty rows and columns."""
        # Produce the mapping between the two tilings
        if not self.active_cells:
            assert GriddedPerm.empty_perm() not in self.obstructions
            self._cached_properties["forward_map"] = RowColMap.identity((0, 0))
            self._obstructions = (GriddedPerm.single_cell((0,), (0, 0)),)
            self._requirements = tuple()
            self._assumptions = tuple()
            self._cached_properties["dimensions"] = (1, 1)
            return
        forward_map = self._minimize_mapping()
        self._cached_properties["forward_map"] = forward_map
        # We still may need to remove point obstructions if the empty row or col
        # was on the end so we do it outside the next if statement.
        self._obstructions = tuple(
            forward_map.map_gp(ob)
            for ob in self.obstructions
            if not ob.is_point_perm() or forward_map.is_mappable_gp(ob)
        )

        if not forward_map.is_identity():
            self._requirements = tuple(
                tuple(forward_map.map_gp(req) for req in reqlist)
                for reqlist in self._requirements
            )
            self._assumptions = tuple(
                sorted(
                    forward_map.map_assumption(assumption)
                    for assumption in self._assumptions
                )
            )
            self._cached_properties["active_cells"] = frozenset(
                forward_map.map_cell(cell)
                for cell in self._cached_properties["active_cells"]
                if forward_map.is_mappable_cell(cell)
            )
            self._cached_properties["empty_cells"] = frozenset(
                forward_map.map_cell(cell)
                for cell in self._cached_properties["empty_cells"]
                if forward_map.is_mappable_cell(cell)
            )
            self._cached_properties["dimensions"] = (
                forward_map.max_col() + 1,
                forward_map.max_row() + 1,
            )

    def _minimize_mapping(self) -> RowColMap:
        """
        Returns a pair of dictionaries, that map rows/columns to an
        equivalent set of rows/columns where empty ones have been removed.
        Also returns a boolean describing whether this mapping is the identity
        mapping which saves some later computation.
        """
        active_cells = self.active_cells
        assert active_cells
        col_set = set(c[0] for c in active_cells)
        row_set = set(c[1] for c in active_cells)
        col_list, row_list = sorted(col_set), sorted(row_set)
        identity = (self.dimensions[0] == len(col_list)) and (
            self.dimensions[1] == len(row_list)
        )
        col_mapping = {x: actual for actual, x in enumerate(col_list)}
        row_mapping = {y: actual for actual, y in enumerate(row_list)}
        return RowColMap(row_map=row_mapping, col_map=col_mapping, is_identity=identity)

    def clean_assumptions(self) -> None:
        """
        Clean assumptions with respect to the known obstructions.

        TODO: this should remove points that are placed, and other requirements
        that are contained in every gridded perm.
        """
        res: List[TrackingAssumption] = []
        for assumption in self.assumptions:
            ass = assumption.avoiding(self._obstructions, self.active_cells)
            if ass.gps:
                res.append(ass)
        self._assumptions = tuple(sorted(set(res)))

    @classmethod
    def guess_from_gridded_perms(
        cls, gps: Iterable[GriddedPerm], max_len: int = -1
    ) -> "Tiling":
        """Given a collection of gridded permutations, attempt to find the Tilings T
        such that the collection of gridded permutations belong to Grid(T). The method
        only looks for obstructions.

        Raises:
        ValueError: If the collection does not contain a pattern that is not avoided
        by all.
        """
        return cls(
            obstructions=guess_obstructions(gps, max_len),
            requirements=(),
            assumptions=(),
        )

    def generate_known_equinumerous_tilings(self) -> Set["Tiling"]:
        """Generate all tilings from known equinumerous mappings."""
        stack, visited = [self], {self}
        while stack:
            for neighbor in Tiling._equinumerous_transpositions(stack.pop()):
                if neighbor not in visited:
                    visited.add(neighbor)
                    stack.append(neighbor)
        return visited

    @staticmethod
    def _equinumerous_transpositions(tiling: "Tiling") -> Iterator["Tiling"]:
        columns, rows = tiling.dimensions
        yield from tiling.all_symmetries()
        yield from (tiling.column_reverse(c) for c in range(columns))
        yield from (tiling.permute_columns(perm) for perm in Perm.of_length(columns))
        for cell in (
            (c, r)
            for c in range(columns)
            for r in range(rows)
            if tiling.contains_all_patterns_locally_for_crossing((c, r))
            and (
                rows == 1
                or all(
                    tiling.is_empty_cell((c, _r))
                    for _r in chain(range(0, r - 1), range(r + 1, rows))
                )
            )
        ):
            yield from Tiling._apply_all_symmetries_to_cell(tiling, cell)
        # Add other pre-existing here:

    @staticmethod
    def _apply_all_symmetries_to_cell(
        tiling: "Tiling", cell: Cell
    ) -> Iterator["Tiling"]:
        yield from (
            tiling.apply_perm_map_to_cell(mapping, cell)
            for mapping in (
                lambda perm: perm.rotate(),
                lambda perm: perm.rotate().inverse(),
                lambda perm: perm.rotate(2),
                lambda perm: perm.rotate(2).inverse(),
                lambda perm: perm.rotate(3),
                lambda perm: perm.rotate(3).inverse(),
            )
        )

    # -------------------------------------------------------------
    # Compression
    # -------------------------------------------------------------

    def to_bytes(self) -> bytes:
        """Compresses the tiling by flattening the sets of cells into lists of
        integers which are concatenated together, every list preceeded by its
        size. The obstructions are compressed and concatenated to the list, as
        are the requirement lists."""

        def split_16bit(n) -> Tuple[int, int]:
            """
            Takes a 16 bit integer and splits it into
            (lower 8bits, upper 8bits).
            """
            return (n & 0xFF, (n >> 8) & 0xFF)

        result = []  # type: List[int]
        result.extend(split_16bit(len(self.obstructions)))
        result.extend(
            chain.from_iterable([len(ob)] + ob.compress() for ob in self.obstructions)
        )
        result.extend(split_16bit(len(self.requirements)))
        for reqlist in self.requirements:
            result.extend(split_16bit(len(reqlist)))
            result.extend(
                chain.from_iterable([len(req)] + req.compress() for req in reqlist)
            )
        if self.assumptions:
            result.extend(split_16bit(len(self.assumptions)))
            for assumption in self.assumptions:
                if isinstance(assumption, SkewComponentAssumption):
                    result.append(2)
                elif isinstance(assumption, SumComponentAssumption):
                    result.append(1)
                elif isinstance(assumption, TrackingAssumption):
                    result.append(0)
                else:
                    raise ValueError("Not a valid assumption.")
                result.extend(split_16bit(len(assumption.gps)))
                result.extend(
                    chain.from_iterable(
                        [len(gp)] + gp.compress() for gp in assumption.gps
                    )
                )
        res = array("B", result)
        return res.tobytes()

    @classmethod
    def from_bytes(cls, b: bytes) -> "Tiling":
        """Given a compressed tiling in the form of an 1-byte array, decompress
        it and return a tiling."""
        # pylint: disable=too-many-locals

        def merge_8bit(lh, uh):
            """
            Takes two 16 bit integers and merges them into
            one 16 bit integer assuming lh is lower half and
            uh is the upper half.
            """
            return lh | (uh << 8)

        def recreate_gp_list(offset):
            """
            Return the gplist implied started at the offset, together
            with the offset after reading it.
            """
            ngps = merge_8bit(arr[offset], arr[offset + 1])
            offset += 2
            res = []
            for _ in range(ngps):
                pattlen = arr[offset]
                offset += 1
                res.append(GriddedPerm.decompress(arr[offset : offset + 3 * pattlen]))
                offset += 3 * pattlen
            return res, offset

        arr = array("B", b)
        obstructions, offset = recreate_gp_list(0)

        nreqs = merge_8bit(arr[offset], arr[offset + 1])
        offset += 2
        requirements = []
        for _ in range(nreqs):
            reqlist, offset = recreate_gp_list(offset)
            requirements.append(reqlist)

        assumptions = []
        if offset < len(arr):
            nassumptions = merge_8bit(arr[offset], arr[offset + 1])
            offset += 2
            for _ in range(nassumptions):
                assumption_type = arr[offset]
                offset += 1
                gps, offset = recreate_gp_list(offset)
                if assumption_type == 0:
                    # tracking
                    assumptions.append(TrackingAssumption(gps))
                elif assumption_type == 1:
                    # sum
                    assumptions.append(SumComponentAssumption(gps))
                elif assumption_type == 2:
                    # skew
                    assumptions.append(SkewComponentAssumption(gps))
                else:
                    raise ValueError("Invalid assumption type.")
        return cls(
            obstructions=obstructions,
            requirements=requirements,
            assumptions=assumptions,
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        )

    @classmethod
    def from_string(cls, string: str) -> "Tiling":
        """Return a 1x1 tiling from string of form 'p1_p2'"""
        basis = [
            GriddedPerm.single_cell(Perm.to_standard(p), (0, 0))
            for p in string.split("_")
        ]
        return cls(obstructions=basis)

    # -------------------------------------------------------------
    # JSON methods
    # -------------------------------------------------------------

    def to_jsonable(self) -> dict:
        """Returns a dictionary object which is JSON serializable which
        represents a Tiling."""
        output: dict = super().to_jsonable()
        output["obstructions"] = [gp.to_jsonable() for gp in self.obstructions]
        output["requirements"] = [
            [gp.to_jsonable() for gp in req] for req in self.requirements
        ]
        output["assumptions"] = [ass.to_jsonable() for ass in self.assumptions]
        return output

    @classmethod
    def from_json(cls, jsonstr: str) -> "Tiling":
        """Returns a Tiling object from JSON string."""
        jsondict = json.loads(jsonstr)
        return cls.from_dict(jsondict)

    @classmethod
    def from_dict(cls, d: dict) -> "Tiling":
        """Returns a Tiling object from a dictionary loaded from a JSON
        serialized Tiling object."""
        obstructions = map(GriddedPerm.from_dict, d["obstructions"])
        requirements = map(lambda x: map(GriddedPerm.from_dict, x), d["requirements"])
        assumptions = map(TrackingAssumption.from_dict, d.get("assumptions", []))
        return cls(
            obstructions=obstructions,
            requirements=requirements,
            assumptions=assumptions,
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        )

    # -------------------------------------------------------------
    # Cell methods
    # -------------------------------------------------------------

    def cell_within_bounds(self, cell: Cell) -> bool:
        """Checks if a cell is within the bounds of the tiling."""
        (i, j) = self.dimensions
        return cell[0] >= 0 and cell[0] < i and cell[1] >= 0 and cell[1] < j

    def empty_cell(self, cell: Cell) -> "Tiling":
        """Empties a cell in the tiling by adding a point obstruction into the
        cell.
        """
        if not self.cell_within_bounds(cell):
            raise ValueError(f"Cell {cell} is not within the bounds of the tiling.")
        return self.add_single_cell_obstruction(Perm((0,)), cell)

    def insert_cell(self, cell: Cell) -> "Tiling":
        """Performs 'cell insertion', adds a point requirement into the given
        cell. Cell should be active.
        """
        if not self.cell_within_bounds(cell):
            raise ValueError(f"Cell {cell} is not within the bounds of the tiling.")
        return self.add_single_cell_requirement(Perm((0,)), cell)

    def add_obstruction(self, patt: Perm, pos: Iterable[Cell]) -> "Tiling":
        """Returns a new tiling with the obstruction of the pattern
        patt with positions pos."""
        return self.add_obstructions((GriddedPerm(patt, pos),))

    def add_obstructions(self, gps: Iterable[GriddedPerm]) -> "Tiling":
        """Returns a new tiling with the obstructions added."""
        new_obs = tuple(gps)
        return Tiling(
            sorted(self._obstructions + new_obs),
            self._requirements,
            self._assumptions,
            sorted_input=True,
            derive_empty=False,
        )

    def add_list_requirement(self, req_list: Iterable[GriddedPerm]) -> "Tiling":
        """
        Return a new tiling with the requirement list added.
        """
        new_req = tuple(sorted(req_list))
        return Tiling(
            self._obstructions,
            sorted(self._requirements + (new_req,)),
            self._assumptions,
            sorted_input=True,
            already_minimized_obs=True,
            derive_empty=False,
        )

    def add_requirement(self, patt: Perm, pos: Iterable[Cell]) -> "Tiling":
        """Returns a new tiling with the requirement of the pattern
        patt with position pos."""
        new_req_list = (GriddedPerm(patt, pos),)
        return self.add_list_requirement(new_req_list)

    def add_single_cell_obstruction(self, patt: Perm, cell: Cell) -> "Tiling":
        """Returns a new tiling with the single cell obstruction of the pattern
        patt in the given cell."""
        return self.add_obstructions((GriddedPerm.single_cell(patt, cell),))

    def add_single_cell_requirement(self, patt: Perm, cell: Cell) -> "Tiling":
        """Returns a new tiling with the single cell requirement of the pattern
        patt in the given cell."""
        new_req_list = (GriddedPerm.single_cell(patt, cell),)
        return self.add_list_requirement(new_req_list)

    def remove_requirement(self, requirement: Tuple[GriddedPerm, ...]) -> "Tiling":
        """Return the tiling where the requirement is removed"""
        try:
            idx = self._requirements.index(requirement)
        except ValueError as e:
            raise ValueError(
                f"following requirement not on tiling: {requirement}"
            ) from e
        return Tiling(
            self._obstructions,
            self._requirements[:idx] + self._requirements[idx + 1 :],
            self._assumptions,
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        )

    def add_assumption(self, assumption: TrackingAssumption) -> "Tiling":
        """Returns a new tiling with the added assumption."""
        return self.add_assumptions((assumption,))

    def add_assumptions(
        self, assumptions: Iterable[TrackingAssumption], clean: bool = True
    ) -> "Tiling":
        """Returns a new tiling with the added assumptions."""
        tiling = Tiling(
            self._obstructions,
            self._requirements,
            self._assumptions + tuple(assumptions),
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        )
        if clean:
            tiling.clean_assumptions()
        return tiling

    def remove_assumption(self, assumption: TrackingAssumption):
        """Returns a new tiling with assumption removed."""
        try:
            idx = self._assumptions.index(assumption)
        except ValueError as e:
            raise ValueError(
                f"following assumption not on tiling: '{assumption}'"
            ) from e
        tiling = Tiling(
            self._obstructions,
            self._requirements,
            self._assumptions[:idx] + self._assumptions[idx + 1 :],
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        )
        tiling.clean_assumptions()
        return tiling

    def remove_assumptions(self) -> "Tiling":
        """
        Return the tiling with all assumptions removed.
        """
        return self.__class__(
            self._obstructions,
            self._requirements,
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        )

    def remove_components_from_assumptions(self):
        """
        Return the tiling with all the actual components from individual
        assumptions removed.
        """
        if not self.assumptions:
            return self
        assumptions = [ass.remove_components(self) for ass in self.assumptions]
        tiling = self.__class__(
            self._obstructions,
            self._requirements,
            [ass for ass in assumptions if ass.gps],
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        )
        tiling.clean_assumptions()
        return tiling

    def fully_isolated(self) -> bool:
        """Check if all cells are isolated on their rows and columns."""
        seen_row: List[int] = []
        seen_col: List[int] = []
        for i, j in self.active_cells:
            if i in seen_col or j in seen_row:
                return False
            seen_col.append(i)
            seen_row.append(j)
        return True

    def only_positive_in_row_and_col(self, cell: Cell) -> bool:
        """Check if the cell is the only positive cell in row and column."""
        if cell not in self.positive_cells:
            return False
        return (
            sum(1 for (x, y) in self.positive_cells if y == cell[1] or x == cell[0])
            == 1
        )

    def only_positive_in_row(self, cell: Cell) -> bool:
        """Check if the cell is the only positive cell in row."""
        if cell not in self.positive_cells:
            return False
        inrow = sum(1 for (x, y) in self.positive_cells if y == cell[1])
        return inrow == 1

    def only_positive_in_col(self, cell: Cell) -> bool:
        """Check if the cell is the only positive cell in column."""
        if cell not in self.positive_cells:
            return False
        incol = sum(1 for (x, y) in self.positive_cells if x == cell[0])
        return incol == 1

    def only_cell_in_col(self, cell: Cell) -> bool:
        """Checks if the cell is the only active cell in the column."""
        return sum(1 for (x, y) in self.active_cells if x == cell[0]) == 1

    def only_cell_in_row(self, cell: Cell) -> bool:
        """Checks if the cell is the only active cell in the row."""
        return sum(1 for (x, y) in self.active_cells if y == cell[1]) == 1

    def only_cell_in_row_and_col(self, cell: Cell) -> bool:
        """Checks if the cell is the only active cell in the row and column."""
        return (
            sum(1 for (x, y) in self.active_cells if y == cell[1] or x == cell[0]) == 1
        )

    def cells_in_row(self, row: int) -> CellFrozenSet:
        """Return all active cells in row."""
        return frozenset((x, y) for (x, y) in self.active_cells if y == row)

    def cells_in_col(self, col: int) -> CellFrozenSet:
        """Return all active cells in column."""
        return frozenset((x, y) for (x, y) in self.active_cells if x == col)

    def cell_basis(self) -> Dict[Cell, Tuple[List[Perm], List[Perm]]]:
        """Returns a dictionary from cells to basis.

        The basis for each cell is a tuple of two lists of permutations.  The
        first list contains the patterns of the obstructions localized in the
        cell and the second contains the intersections of requirement lists
        that are localized in the cell.
        """
        try:
            return self._cached_properties["cell_basis"]
        except KeyError:
            obdict: Dict[Cell, List[Perm]] = defaultdict(list)
            reqdict: Dict[Cell, List[Perm]] = defaultdict(list)
            for ob in self.obstructions:
                if ob.is_localized():
                    cell = ob.pos[0]
                    obdict[cell].append(ob.patt)

            for req_list in self.requirements:
                for gp in req_list:
                    for cell in set(gp.pos):
                        subgp = gp.get_gridded_perm_in_cells([cell])
                        if subgp not in reqdict[cell] and all(
                            subgp in r for r in req_list
                        ):
                            reqdict[cell].append(subgp.patt)
            for cell, contain in reqdict.items():
                ind_to_remove = set()
                for i, req in enumerate(contain):
                    if any(req in other for j, other in enumerate(contain) if i != j):
                        ind_to_remove.add(i)
                reqdict[cell] = [
                    req for i, req in enumerate(contain) if i not in ind_to_remove
                ]

            all_cells = product(range(self.dimensions[0]), range(self.dimensions[1]))
            resdict = {cell: (obdict[cell], reqdict[cell]) for cell in all_cells}
            self._cached_properties["cell_basis"] = resdict
            return resdict

    def cell_graph(self) -> Set[Tuple[Cell, Cell]]:
        """
        Return the set of edges in the cell graph of the tiling.
        """
        edges = set()
        cells = sorted(self.active_cells)
        for c1, c2 in zip(cells[:-1], cells[1:]):
            if c1[0] == c2[0]:
                edges.add((c1, c2))
        cells = sorted(self.active_cells, key=lambda x: (x[1], x[0]))
        for c1, c2 in zip(cells[:-1], cells[1:]):
            if c1[1] == c2[1]:
                edges.add((c1, c2))
        return edges

    def sum_decomposition(self, skew: bool = False) -> List[List[Cell]]:
        """
        Returns the sum decomposition of the tiling with respect to the cells.
        If skew is True then returns the skew decomposition instead.
        """
        cells = sorted(self.active_cells)
        decomposition: List[List[Cell]] = []
        while len(cells) > 0:
            x = cells[0][0]  # x boundary, maximum in both cases
            y = cells[0][1]  # y boundary, maximum in sum, minimum in skew
            change = True
            while change:
                change = False
                for c in cells:
                    if c[0] <= x:
                        if (skew and c[1] < y) or (not skew and c[1] > y):
                            y = c[1]
                            change = True
                    if (skew and c[1] >= y) or (not skew and c[1] <= y):
                        if c[0] > x:
                            x = c[0]
                            change = True
            decomposition.append([])
            new_cells = []
            for c in cells:
                if c[0] <= x:
                    decomposition[-1].append(c)
                else:
                    new_cells.append(c)
            cells = new_cells
        return decomposition

    def skew_decomposition(self) -> List[List[Cell]]:
        """
        Returns the skew decomposition of the tiling with respect to the cells
        """
        return self.sum_decomposition(skew=True)

    @staticmethod
    def sort_requirements(
        requirements: Iterable[Iterable[GriddedPerm]],
    ) -> Tuple[Tuple[GriddedPerm, ...], ...]:
        return tuple(sorted(tuple(sorted(set(reqlist))) for reqlist in requirements))

    @property
    def forward_map(self) -> RowColMap:
        try:
            return self._cached_properties["forward_map"]
        except KeyError:
            self._remove_empty_rows_and_cols()
            return self._cached_properties["forward_map"]

    @property
    def backward_map(self) -> RowColMap:
        try:
            return self._cached_properties["backward_map"]
        except KeyError:
            backward_map = self.forward_map.reverse()
            self._cached_properties["backward_map"] = backward_map
            return backward_map

    # -------------------------------------------------------------
    # Symmetries
    # -------------------------------------------------------------

    def _transform(
        self, transf, gptransf: Callable[[GriddedPerm], GriddedPerm]
    ) -> "Tiling":
        """
        Transforms the tiling according to the two transformation functions
        given. The first transf is mapping of cells while gptransf is a
        transformation of GriddedPerm that calls some internal method.
        # TODO: transf is not used...
        """
        return Tiling(
            obstructions=(gptransf(ob) for ob in self.obstructions),
            requirements=(
                [gptransf(req) for req in reqlist] for reqlist in self.requirements
            ),
            assumptions=(
                ass.__class__(gptransf(gp) for gp in ass.gps)
                for ass in self._assumptions
            ),
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
        )

    def reverse(self, regions=False):
        """
        Reverses the tiling within its boundary. Every cell and obstruction
        gets flipped over the vertical middle axis.
        # TODO: remove weird regions flag?
        """

        def reverse_cell(cell: Cell) -> Cell:
            return (self.dimensions[0] - cell[0] - 1, cell[1])

        reversed_tiling = self._transform(
            reverse_cell, lambda gp: gp.reverse(reverse_cell)
        )
        if not regions:
            return reversed_tiling
        return (
            [reversed_tiling],
            [{c: frozenset([reverse_cell(c)]) for c in self.active_cells}],
        )

    def complement(self) -> "Tiling":
        """Flip over the horizontal axis."""

        def complement_cell(cell: Cell) -> Cell:
            return (cell[0], self.dimensions[1] - cell[1] - 1)

        return self._transform(
            complement_cell, lambda gp: gp.complement(complement_cell)
        )

    def inverse(self) -> "Tiling":
        """Flip over the diagonal"""

        def inverse_cell(cell: Cell) -> Cell:
            return (cell[1], cell[0])

        return self._transform(inverse_cell, lambda gp: gp.inverse(inverse_cell))

    def antidiagonal(self) -> "Tiling":
        """Flip over the anti-diagonal"""

        def antidiagonal_cell(cell: Cell) -> Cell:
            return (self.dimensions[1] - cell[1] - 1, self.dimensions[0] - cell[0] - 1)

        return self._transform(
            antidiagonal_cell, lambda gp: gp.antidiagonal(antidiagonal_cell)
        )

    def rotate270(self) -> "Tiling":
        """Rotate 270 degrees"""

        def rotate270_cell(cell: Cell) -> Cell:
            return (self.dimensions[1] - cell[1] - 1, cell[0])

        return self._transform(rotate270_cell, lambda gp: gp.rotate270(rotate270_cell))

    def rotate180(self) -> "Tiling":
        """Rotate 180 degrees"""

        def rotate180_cell(cell: Cell) -> Cell:
            return (self.dimensions[0] - cell[0] - 1, self.dimensions[1] - cell[1] - 1)

        return self._transform(rotate180_cell, lambda gp: gp.rotate180(rotate180_cell))

    def rotate90(self) -> "Tiling":
        """Rotate 90 degrees"""

        def rotate90_cell(cell: Cell) -> Cell:
            return (cell[1], self.dimensions[0] - cell[0] - 1)

        return self._transform(rotate90_cell, lambda gp: gp.rotate90(rotate90_cell))

    def all_symmetries(self) -> Set["Tiling"]:
        """
        Return all the symmetries of a tiling in a set.
        """
        symmetries = set()
        t = self
        for _ in range(4):
            symmetries.add(t)
            symmetries.add(t.inverse())
            t = t.rotate90()
        return symmetries

    def column_reverse(self, column: int) -> "Tiling":
        """Reverse a given column in the tiling."""
        return self._transform(lambda _c: _c, lambda gp: gp.column_reverse(column))

    def row_complement(self, row: int) -> "Tiling":
        """Changes a row to its complement."""
        return self._transform(lambda _c: _c, lambda gp: gp.row_complement(row))

    def permute_columns(self, perm: Iterable[int]) -> "Tiling":
        """Given an initial state of columns 12...n, permute them using the provided
        permutation.."""
        if not isinstance(perm, Perm):
            perm = Perm(perm)
        assert len(perm) == self.dimensions[0]
        return self._transform(lambda _c: _c, lambda gp: gp.permute_columns(perm))

    def permute_rows(self, perm: Iterable[int]) -> "Tiling":
        """Given an initial state of rows 12...n, permute them using the provided
        permutation.."""
        if not isinstance(perm, Perm):
            perm = Perm(perm)
        assert len(perm) == self.dimensions[1]
        return self._transform(lambda _c: _c, lambda gp: gp.permute_rows(perm))

    def apply_perm_map_to_cell(
        self,
        perm_mapping: Callable[[Perm], Perm],
        cell: Cell,
    ) -> "Tiling":
        """Apply a permutation mapping on anything within a cell."""
        return self._transform(
            lambda _c: _c, lambda gp: gp.apply_perm_map_to_cell(perm_mapping, cell)
        )

    # -------------------------------------------------------------
    # Algorithms
    # -------------------------------------------------------------

    def _fusion(self, row, col, fusion_class):
        """
        Fuse the tilings using the fusion class.

        If `row` is not `None` then `row` and `row+1` are fused together.
        If `col` is not `None` then `col` and `col+1` are fused together.
        """
        assert xor(row is None, col is None), "Specify only `row` or `col`"
        if not (
            row in range(self.dimensions[1] - 1) or col in range(self.dimensions[0] - 1)
        ):
            raise InvalidOperationError("`row` or `column` out or range")
        fusion = fusion_class(self, row_idx=row, col_idx=col)
        if not fusion.fusable():
            fus_type = "Rows" if row is not None else "Columns"
            idx = row if row is not None else col
            message = f"{fus_type} {idx} and {idx+1} are not fusable."
            raise InvalidOperationError(message)
        return fusion.fused_tiling()

    def fusion(self, row=None, col=None):
        """
        Fuse the tilings.

        If `row` is not `None` then `row` and `row+1` are fused together.
        If `col` is not `None` then `col` and `col+1` are fused together.
        """
        return self._fusion(row, col, Fusion)

    def component_fusion(self, row=None, col=None):
        """
        Fuse the tilings in such a way that it can be unfused by drawing a line
        between skew/sum-components.

        If `row` is not `None` then `row` and `row+1` are fused together.
        If `col` is not `None` then `col` and `col+1` are fused together.
        """
        return self._fusion(row, col, ComponentFusion)

    def sub_tiling(
        self,
        cells: Iterable[Cell],
        factors: bool = False,
        add_assumptions: Iterable[TrackingAssumption] = tuple(),
    ) -> "Tiling":
        """Return the tiling using only the obstructions and requirements
        completely contained in the given cells. If factors is set to True,
        then it assumes that the first cells confirms if a gridded perm uses only
        the cells."""
        obstructions = tuple(
            ob
            for ob in self.obstructions
            if (factors and ob.pos[0] in cells) or all(c in cells for c in ob.pos)
        )
        requirements = Tiling.sort_requirements(
            req
            for req in self.requirements
            if (factors and req[0].pos[0] in cells)
            or all(c in cells for c in chain.from_iterable(r.pos for r in req))
        )
        assumptions = tuple(
            ass.__class__(
                gp
                for gp in ass.gps
                if (factors and gp.pos[0] in cells) or all(c in cells for c in gp.pos)
            )
            for ass in self.assumptions
        ) + tuple(add_assumptions)
        # TODO: check sum/skew assumptions
        return self.__class__(
            obstructions,
            requirements,
            tuple(sorted(set(ass for ass in assumptions if ass.gps))),
            simplify=False,
            sorted_input=True,
        )

    def find_factors(self, interleaving: str = "none") -> Tuple["Tiling", ...]:
        """
        Return list with the factors of the tiling.

        Two non-empty cells are in the same factor if they are in the same row
        or column, or they share an obstruction or requirement. However, if
        `interleaving` is set to 'monotone' cell on the same row do not need to
        be in the same factor if one of them is monotone. If interleaving is
        set 'all' then cells on the same row or columns don't need to be in
        the same factor.
        """
        factor_class = {
            "none": Factor,
            "monotone": FactorWithMonotoneInterleaving,
            "any": FactorWithInterleaving,
        }
        if interleaving in factor_class:
            factor = factor_class[interleaving](self)
        else:
            raise InvalidOperationError(
                f"interleaving option must be in {list(factor_class.keys())}"
            )
        return factor.factors()

    def row_and_column_separation(self) -> "Tiling":
        """
        Splits the row and columns of a tilings using the inequalities implied
        by the length two obstructions.
        """
        rcs = RowColSeparation(self)
        return rcs.separated_tiling()

    def row_and_column_separation_with_mapping(
        self,
    ) -> Tuple["Tiling", Dict[Cell, Cell]]:
        rcs = RowColSeparation(self)
        return rcs.separated_tiling(), rcs.get_cell_map()

    def obstruction_transitivity(self) -> "Tiling":
        """
        Add length 2 obstructions to the tiling using transitivity over
        positive cells.

        For three cells A, B and C on the same row or column, if A < B, B < C
        and B is positive then the obstruction for A < C is added.
        """
        obs_trans = ObstructionTransitivity(self)
        return obs_trans.obstruction_transitivity()

    def all_obstruction_inferral(self, obstruction_length: int) -> "Tiling":
        """
        Adds all the obstruction of length up to `obstruction_length` that
        does not change the set of gridded permutations of the tiling.
        """
        obs_inf = AllObstructionInferral(self, obstruction_length)
        return obs_inf.obstruction_inferral()

    def empty_cell_inferral(self) -> "Tiling":
        """
        Adds point obstruction in the cell of the tilings that should be empty.

        Equivalent to `self.all_obstruction_inferral(1)`
        """
        obs_inf = EmptyCellInferral(self)
        return obs_inf.obstruction_inferral()

    def subobstruction_inferral(self) -> "Tiling":
        """
        Adds all the subobstruction of the tiling's obstruction that doesn't
        change the set of gridded permutations of the tiling.
        """
        obs_inf = SubobstructionInferral(self)
        return obs_inf.obstruction_inferral()

    def place_point_in_cell(self, cell: Cell, direction: int) -> "Tiling":
        """
        Return the tiling where a point is placed in the given cell and
        direction.
        """
        req_placement = RequirementPlacement(self)
        return req_placement.place_point_in_cell(cell, direction)

    def place_point_of_gridded_permutation(
        self, gp: GriddedPerm, idx: int, direction: int
    ) -> "Tiling":
        """
        Return the tiling where the directionmost occurrence of the idx point
        in the gridded permutaion gp is placed.
        """
        req_placement = RequirementPlacement(self)
        return req_placement.place_point_of_gridded_permutation(gp, idx, direction)

    def place_row(self, idx: int, direction: int) -> Tuple["Tiling", ...]:
        """
        Return the list of tilings in which the directionmost point in the row
        is placed.
        """
        req_placement = RequirementPlacement(self)
        return req_placement.row_placement(idx, direction)

    def place_col(self, idx: int, direction: int) -> Tuple["Tiling", ...]:
        """
        Return the list of tilings in which the directionmost point in the
        column is placed.
        """
        req_placement = RequirementPlacement(self)
        return req_placement.col_placement(idx, direction)

    def partial_place_point_in_cell(self, cell: Cell, direction: int) -> "Tiling":
        """
        Return the tiling where a point is placed in the given cell and
        direction. The point is placed onto its own row or own column
        depending on the direction.
        """
        if direction in (DIR_EAST, DIR_WEST):
            req_placement = RequirementPlacement(self, own_row=False)
        else:
            req_placement = RequirementPlacement(self, own_col=False)
        return req_placement.place_point_in_cell(cell, direction)

    def partial_place_point_of_gridded_permutation(
        self, gp: GriddedPerm, idx: int, direction: int
    ) -> "Tiling":
        """
        Return the tiling where the directionmost occurrence of the idx point
        in the gridded permutaion gp is placed. The point is placed onto its
        own row or own column depending on the direction.
        """
        if direction in (DIR_EAST, DIR_WEST):
            req_placement = RequirementPlacement(self, own_row=False)
        else:
            req_placement = RequirementPlacement(self, own_col=False)
        return req_placement.place_point_of_gridded_permutation(gp, idx, direction)

    def partial_place_row(self, idx: int, direction: int) -> Tuple["Tiling", ...]:
        """
        Return the list of tilings in which the directionmost point in the row
        is placed. The points are placed onto thier own row.
        """
        req_placement = RequirementPlacement(self, own_row=True, own_col=False)
        return req_placement.row_placement(idx, direction)

    def partial_place_col(self, idx: int, direction: int) -> Tuple["Tiling", ...]:
        """
        Return the list of tilings in which the directionmost point in the
        column is placed. The points are placed onto their own column.
        """
        req_placement = RequirementPlacement(self, own_row=False, own_col=True)
        return req_placement.col_placement(idx, direction)

    def is_subclass(self, perms_to_check: Iterable[Perm]):
        """
        Returns true if the underlying permutations of all the gridded permutations
        on the tiling lie within one of the principal classes whose basis elements
        are given in perms_to_check.
        """
        return SubclassVerificationAlgorithm(self, set(perms_to_check)).is_verified()

    # -------------------------------------------------------------
    # HTML methods
    # -------------------------------------------------------------

    def _handle_html_assumption(self, result: List[str], style) -> List[str]:
        """adds background color in cells where assumption happens"""
        # pylint: disable=too-many-locals
        colors = [
            "#b0dbff",
            "#d1f0af",
            "#db8686",
            "#FCC997",
            "#b0ffd0",
            "#FCEB97",
            "#fc97b4",
            "#4b45ff",
            "#c8bdff",
            "#bfbfbf",
        ]
        has_ass: Dict[int, List[str]] = {}
        for c, ass in enumerate(self.assumptions):
            for gp in ass.gps:
                if len(gp.pos) > 1:
                    pass
                else:
                    i, j = gp.pos[0]
                    dim_i, dim_j = self.dimensions
                    index = (dim_j - j - 1) * (3 * dim_i + 2) + i * 3 + 2
                    if c >= len(colors):
                        pass
                    elif index in has_ass:
                        has_ass[index].append(colors[c])
                    else:
                        has_ass[index] = [colors[c]]

                    if c >= len(colors) or len(has_ass[index]) > 4:
                        # display gray lines if out of color or
                        # more than 4 assumption in single cell
                        background_image = """background-image:
                            repeating-linear-gradient(
                            45deg, #ffffff, #ffffff 6px, #00000080 1px, #00000080 7px
                            );"""
                    else:
                        # display stripes
                        background_image = "background-image: linear-gradient(180deg"
                        stripe_size = 24 // len(has_ass[index])
                        for i, color in enumerate(has_ass[index]):
                            background_image += f""",
                                {color} {i*stripe_size}px,
                                {color} {(i+1)*stripe_size}px"""
                        background_image += ");"
                    result[index] = f'<th style="{background_image}{style}">'
        return result

    def to_html_representation(self) -> str:
        """Returns an html representation of the tilings object"""
        # pylint: disable=too-many-locals
        # stylesheet for tiling
        style = """
            border: 1px solid;
            width: 24px;
            height: 24px;
            text-align: center;
            """
        dim_i, dim_j = self.dimensions
        result = []
        # Create tiling html table
        result.append("<table> ")
        for _ in range(dim_j):
            result.append("<tr>")
            for _ in range(dim_i):
                result.append(f"<th style='{style}'>")
                result.append(" ")
                result.append("</th>")
            result.append("</tr>")
        result.append("</table>")
        labels: Dict[Tuple[Tuple[Perm, ...], bool], str] = {}

        # Put the sets in the tiles

        # How many characters are in a row in the grid
        row_width = 3 * dim_i + 2
        curr_label = 1
        for cell, gridded_perms in sorted(self.cell_basis().items()):
            obstructions, _ = gridded_perms
            basis = list(sorted(obstructions))
            if basis == [Perm((0,))]:
                continue
            # the block, is the basis and whether or not positive
            block = (tuple(basis), cell in self.positive_cells)
            label = labels.get(block)
            if label is None:
                if basis == [Perm((0, 1)), Perm((1, 0))]:
                    if cell in self.positive_cells:
                        label = "\u25cf"
                    else:
                        label = "\u25cb"
                elif basis == [Perm((0, 1))]:
                    label = "\\"
                elif basis == [Perm((1, 0))]:
                    label = "/"
                else:
                    label = str(curr_label)
                    curr_label += 1
                labels[block] = label
            row_index_from_top = dim_j - cell[1] - 1
            index = row_index_from_top * row_width + cell[0] * 3 + 3
            result[index] = label

        # adds background color in cells where assumption happens
        result = self._handle_html_assumption(result, style)
        return "".join(result)

    # -------------------------------------------------------------
    # Properties and getters
    # -------------------------------------------------------------

    @property
    def extra_parameters(self) -> Tuple[str, ...]:
        return tuple(f"k_{i}" for i in range(len(self._assumptions)))

    def get_parameters(self, obj: GriddedPerm) -> Parameters:
        return tuple(ass.get_value(obj) for ass in self.assumptions)

    def possible_parameters(self, n: int) -> Iterator[Dict[str, int]]:
        if any(
            len(gp) > 1
            for gp in chain.from_iterable(ass.gps for ass in self.assumptions)
        ):
            raise NotImplementedError(
                "possible parameters only implemented for assumptions with "
                "size one gridded perms"
            )
        parameters = [self.get_assumption_parameter(ass) for ass in self.assumptions]
        for values in product(*[range(n + 1) for _ in parameters]):
            yield dict(zip(parameters, values))

    def get_assumption_parameter(self, assumption: TrackingAssumption) -> str:
        """
        Return the variable associated with the given assumption.

        Raise ValueError if the assumptions is not on the tiling.
        """
        try:
            idx = self._assumptions.index(assumption)
        except ValueError as e:
            raise ValueError(
                f"following assumption not on tiling: '{assumption}'"
            ) from e
        return f"k_{idx}"

    def get_assumption(self, parameter: str) -> TrackingAssumption:
        idx = parameter.split("_")[1]
        return self.assumptions[int(idx)]

    def get_minimum_value(self, parameter: str) -> int:
        """
        Return the minimum value that can be taken by the parameter.
        """
        assumption = self.get_assumption(parameter)
        if isinstance(assumption, ComponentAssumption):
            return (
                1
                if any(gp.pos[0] in self.positive_cells for gp in assumption.gps)
                else 0
            )
        return min(assumption.get_value(gp) for gp in self.minimal_gridded_perms())

    def maximum_length_of_minimum_gridded_perm(self) -> int:
        """Returns the maximum length of the minimum gridded permutation that
        can be gridded on the tiling.
        """
        return sum(max(map(len, reqs)) for reqs in self.requirements)

    def is_empty(self) -> bool:
        """Checks if the tiling is empty.

        Tiling is empty if it has been inferred to be contradictory due to
        contradicting requirements and obstructions or no gridded permutation
        can be gridded on the tiling.
        """
        if any(ob.is_empty() for ob in self.obstructions):
            return True
        if len(self.requirements) <= 1:
            return False
        MGP = MinimalGriddedPerms(self.obstructions, self.requirements)
        return all(False for _ in MGP.minimal_gridded_perms(yield_non_minimal=True))

    def is_finite(self) -> bool:
        """Returns True if all active cells have finite basis."""
        increasing = set()
        decreasing = set()
        for ob in self.obstructions:
            if ob.is_single_cell():
                if ob.patt.is_increasing():
                    increasing.add(ob.pos[0])
                if ob.patt.is_decreasing():
                    decreasing.add(ob.pos[0])
        return all(
            cell in increasing and cell in decreasing for cell in self.active_cells
        )

    def is_increasing(self) -> bool:
        """Returns true if all gridded perms are increasing."""
        separated = self.row_and_column_separation()
        components = separated.sum_decomposition()
        if any(len(cells) > 1 for cells in components):
            return False
        cells = sorted(cells[0] for cells in components)
        if any(b < a for a, b in zip(cells[:-1], cells[1:])):
            return False
        return all(
            GriddedPerm.single_cell(Perm((1, 0)), cell) in separated.obstructions
            for cell in cells
        )

    def is_decreasing(self) -> bool:
        """Returns true if all gridded perms are decreasing."""
        separated = self.row_and_column_separation()
        components = separated.skew_decomposition()
        if any(len(cells) > 1 for cells in components):
            return False
        cells = sorted(cells[0] for cells in components)
        if any(b > a for a, b in zip(cells[:-1], cells[1:])):
            return False
        return all(
            GriddedPerm.single_cell(Perm((0, 1)), cell) in separated.obstructions
            for cell in cells
        )

    def objects_of_size(self, n: int, **parameters: int) -> Iterator[GriddedPerm]:
        for gp in self.gridded_perms_of_length(n):
            if all(
                self.get_assumption(k).get_value(gp) == val
                for k, val in parameters.items()
            ):
                yield gp

    def gridded_perms_of_length(self, length: int) -> Iterator[GriddedPerm]:
        for gp in self.gridded_perms(maxlen=length):
            if len(gp) == length:
                yield gp

    def initial_conditions(self, check: int = 6) -> List[Any]:
        """
        Returns a list with the initial conditions to size `check` of the
        CombinatorialClass.
        """
        res = [0 for _ in range(check + 1)]
        extra_params = self.extra_parameters
        ass_counter = [
            (sympy.var(k), self.get_assumption(k).get_value) for k in extra_params
        ]
        for gp in self.gridded_perms(check):
            res[len(gp)] += reduce(
                mul,
                (var ** func(gp) for var, func in ass_counter),
                sympy.Number(1),
            )
        return res

    def gridded_perms(self, maxlen: Optional[int] = None) -> Iterator[GriddedPerm]:
        """
        Iterator of all gridded permutations griddable on the tiling.

        The gridded permutations are up to length of the longest minimum
        gridded permutations that is griddable on the tiling.
        """
        maxlen = (
            maxlen
            if maxlen is not None
            else self.maximum_length_of_minimum_gridded_perm()
        )
        yield from GriddedPermsOnTiling(self).gridded_perms(maxlen)

    def enmerate_gp_up_to(self, max_length: int) -> List[int]:
        """Count gridded perms of each length up to a max length."""
        cnt = Counter(len(gp) for gp in self.gridded_perms(max_length))
        return [cnt[i] for i in range(max_length + 1)]

    def merge(self) -> "Tiling":
        """Return an equivalent tiling with a single requirement list.
        # TODO: this doesn't work due to minimization on initialising"""
        if len(self.requirements) <= 1:
            return self
        mgps = MinimalGriddedPerms(self.obstructions, self.requirements)
        requirements = tuple(
            GriddedPerm(gp.patt, gp.pos) for gp in mgps.minimal_gridded_perms()
        )
        return self.__class__(self.obstructions, (requirements,), self.assumptions)

    def minimal_gridded_perms(self) -> Iterator[GriddedPerm]:
        """
        An iterator over all minimal gridded permutations.
        """
        MGP = MinimalGriddedPerms(self.obstructions, self.requirements)
        yield from MGP.minimal_gridded_perms()

    def is_epsilon(self) -> bool:
        """Returns True if the generating function for the tiling is 1."""
        return (
            self.dimensions == (1, 1)
            and len(self.obstructions) == 1
            and len(self.obstructions[0]) == 1
        )

    def is_positive(self) -> bool:
        """Returns True if tiling does not contain the empty permutation."""
        return bool(self.requirements)

    def is_point_tiling(self) -> bool:
        """
        Returns True if the only gridded permutations of the tiling is
        1: (0, 0)
        """
        return self.dimensions == (1, 1) and (0, 0) in self.point_cells

    def is_atom(self) -> bool:
        """Return True if tiling is a single gridded permutation."""
        return (
            (self.active_cells == self.point_cells)
            and self.fully_isolated()
            and not any(ob.is_empty() for ob in self.obstructions)
        )

    def minimum_size_of_object(self) -> int:
        """Return the size of the smallest gridded perm contained on the tiling."""
        if not self.requirements:
            return 0
        if len(self.requirements) == 1:
            return min(len(gp) for gp in self.requirements[0])
        return len(next(self.minimal_gridded_perms()))

    def is_point_or_empty_cell(self, cell: Cell) -> bool:
        point_or_empty_obs = (
            GriddedPerm((0, 1), (cell, cell)),
            GriddedPerm((1, 0), (cell, cell)),
        )
        return all(ob in self.obstructions for ob in point_or_empty_obs)

    def is_empty_cell(self, cell: Cell) -> bool:
        """Check if the cell of the tiling is empty."""
        return cell in self.empty_cells

    def is_monotone_cell(self, cell: Cell) -> bool:
        """
        Check if the cell is decreasing or increasing.

        If the cell is empty it is considered as monotone.
        """
        local_obs = self.cell_basis()[cell][0]
        return any(ob in [Perm((0,)), Perm((0, 1)), Perm((1, 0))] for ob in local_obs)

    def contains_all_patterns_locally_for_crossing(self, cell: Cell) -> bool:
        """Check if for all crossing obstructions and requirements through a given cell,
        the tiling contains the same crossing obstructions and requirements but with all
        permutations of its subpattern within the cell.
        """
        return all(
            Tiling._contains_all_subpatterns_in_cell(gp_set, cell)
            for gp_set in chain(
                (
                    {
                        req
                        for req in req_list
                        if not req.is_localized() and req.occupies(cell)
                    }
                    for req_list in self.requirements
                ),
                (
                    {
                        obs
                        for obs in self.obstructions
                        if not obs.is_localized() and obs.occupies(cell)
                    },
                ),
            )
        )

    @staticmethod
    def _contains_all_subpatterns_in_cell(gps: Set[GriddedPerm], cell: Cell) -> bool:
        while gps:
            gp = next(iter(gps))
            perm = tuple(gp.patt[i] for i in gp.points_in_cell(cell))
            back_map = dict(zip(Perm.to_standard(perm), perm))
            for patt in Perm.of_length(len(perm)):
                it = iter(patt)
                gp = GriddedPerm(
                    (back_map[next(it)] if c == cell else val for val, c in gp),
                    gp.pos,
                )
                try:
                    gps.remove(gp)
                except KeyError:
                    return False
        return True

    @property
    def point_cells(self) -> CellFrozenSet:
        try:
            return self._cached_properties["point_cells"]
        except KeyError:
            local_length2_obcells = Counter(
                ob.pos[0]
                for ob in self._obstructions
                if ob.is_localized() and len(ob) == 2
            )
            point_cells = frozenset(
                cell for cell in self.positive_cells if local_length2_obcells[cell] == 2
            )
            self._cached_properties["point_cells"] = point_cells
            return point_cells

    @property
    def total_points(self) -> int:
        return len(self.point_cells)

    @property
    def positive_cells(self) -> CellFrozenSet:
        try:
            return self._cached_properties["positive_cells"]
        except KeyError:
            positive_cells = frozenset(
                union_reduce(
                    intersection_reduce(req.pos for req in reqs)
                    for reqs in self._requirements
                )
            )
            self._cached_properties["positive_cells"] = positive_cells
            return positive_cells

    def total_positive(self) -> int:
        return len(self.positive_cells)

    @property
    def possibly_empty(self) -> CellFrozenSet:
        """Computes the set of possibly empty cells on the tiling."""
        try:
            return self._cached_properties["possibly_empty"]
        except KeyError:
            possibly_empty = self.active_cells - self.positive_cells
            self._cached_properties["possibly_empty"] = possibly_empty
            return possibly_empty

    @property
    def obstructions(self) -> Tuple[GriddedPerm, ...]:
        return self._obstructions

    def total_obstructions(self) -> int:
        return len(self._obstructions)

    @property
    def requirements(self) -> Tuple[Tuple[GriddedPerm, ...], ...]:
        return self._requirements

    def total_requirements(self) -> int:
        return len(self._requirements)

    @property
    def assumptions(self) -> Tuple[TrackingAssumption, ...]:
        return self._assumptions

    def total_assumptions(self) -> int:
        return len(self._assumptions)

    @property
    def empty_cells(self) -> CellFrozenSet:
        """Returns a set of all cells that contain a point obstruction, i.e.,
        are empty.
        """
        try:
            return self._cached_properties["empty_cells"]
        except KeyError:
            self._prepare_properties()
            return self._cached_properties["empty_cells"]

    @property
    def active_cells(self) -> CellFrozenSet:
        """
        Returns a set of all cells that do not contain a point obstruction,
        i.e., not empty.
        """
        try:
            return self._cached_properties["active_cells"]
        except KeyError:
            self._prepare_properties()
            return self._cached_properties["active_cells"]

    @property
    def dimensions(self) -> Dimension:
        try:
            return self._cached_properties["dimensions"]
        except KeyError:
            self._prepare_properties()
            return self._cached_properties["dimensions"]

    def add_obstruction_in_all_ways(self, patt: Perm) -> "Tiling":
        """
        Adds an obstruction of the pattern patt in all possible ways to
        a fully separated (no interleaving rows or columns) tiling t.
        """

        def rec(
            cols: List[List[Cell]],
            p: Perm,
            pos: List[Cell],
            used: Dict[int, Cell],
            i: int,
            j: int,
            res: List[GriddedPerm],
        ) -> None:
            """
            Recursive helper function
            cols: List of columns in increasing order, each column is a list of
            cells
            p: The pattern
            pos: List of the pattern's positions
            used: Dictionary mapping permutation values to cells for pruning
            i: Index in cells
            j: Index in p
            res: Resulting list of obstructions
            """
            if j == len(p):
                res.append(GriddedPerm(p, tuple(x for x in pos)))
            elif i == len(cols):
                return
            else:
                upper = min(v[1] for k, v in used.items() if k > p[j])
                lower = max(v[1] for k, v in used.items() if k < p[j])
                for cell in cols[i]:
                    if lower <= cell[1] <= upper:
                        used[p[j]] = cell
                        pos.append(cell)
                        rec(cols, p, pos, used, i, j + 1, res)
                        pos.pop()
                        del used[p[j]]
                rec(cols, p, pos, used, i + 1, j, res)

        cols: List[List[Cell]] = [[] for i in range(self.dimensions[0])]
        for x in self.active_cells:
            cols[x[0]].append(x)
        used = {-1: (-1, -1), len(patt): self.dimensions}
        pos: List[Cell] = []
        res: List[GriddedPerm] = []
        rec(cols, patt, pos, used, 0, 0, res)
        return Tiling(
            obstructions=sorted(list(self.obstructions) + res),
            requirements=self.requirements,
            assumptions=self.assumptions,
            sorted_input=True,
            derive_empty=False,
        )

    @classmethod
    def tiling_from_perm(cls, p: Perm) -> "Tiling":
        """
        Returns a tiling with point requirements corresponding to the
        permutation 'p'
        """
        return cls(
            requirements=[[GriddedPerm((0,), ((i, p[i]),))] for i in range(len(p))]
        )

    def get_genf(self, *args, **kwargs) -> Any:
        # pylint: disable=import-outside-toplevel
        if self.is_empty():
            return sympy.sympify(0)
        from .strategies import (
            BasicVerificationStrategy,
            InsertionEncodingVerificationStrategy,
            LocallyFactorableVerificationStrategy,
            LocalVerificationStrategy,
            MonotoneTreeVerificationStrategy,
            OneByOneVerificationStrategy,
        )

        enum_stragies: List[VerificationStrategy] = [
            BasicVerificationStrategy(),
            OneByOneVerificationStrategy(),
            LocallyFactorableVerificationStrategy(),
            InsertionEncodingVerificationStrategy(),
            MonotoneTreeVerificationStrategy(),
            LocalVerificationStrategy(),
        ]
        for enum_strat in enum_stragies:
            try:
                return enum_strat.get_genf(self)
            except StrategyDoesNotApply:
                continue
        raise NotImplementedError(
            "We were unable to enumerate this tiling:\n" + str(self)
        )

    def to_gui(self):
        """Try to open the current tiling in the GUI."""
        run_gui(json.dumps(self.to_jsonable(), separators=(",", ":")))

    # -------------------------------------------------------------
    # Dunder methods
    # -------------------------------------------------------------

    def __hash__(self) -> int:
        return (
            hash(self._requirements)
            ^ hash(self._obstructions)
            ^ hash(self._assumptions)
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Tiling):
            return False
        return (
            self.obstructions == other.obstructions
            and self.requirements == other.requirements
            and self.assumptions == other.assumptions
        )

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Tiling):
            return True
        return (
            self.obstructions != other.obstructions
            or self.requirements != other.requirements
            or self.assumptions != other.assumptions
        )

    def __contains__(self, gp: GriddedPerm) -> bool:
        """Test if a gridded permtuaiton is griddable on the given tiling."""
        return gp.avoids(*self.obstructions) and all(
            gp.contains(*req) for req in self.requirements
        )

    def __repr__(self) -> str:
        format_string = "Tiling(obstructions={}, requirements={}, assumptions={})"
        non_point_obstructions = tuple(
            filterfalse(GriddedPerm.is_point_perm, self.obstructions)
        )
        return format_string.format(
            non_point_obstructions, self.requirements, self.assumptions
        )

    def __str__(self) -> str:
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements
        dim_i, dim_j = self.dimensions
        result = []
        # Create tiling lines
        for j in range(2 * dim_j + 1):
            for i in range(2 * dim_i + 1):
                # Whether or not a vertical line and a horizontal line is
                # present
                vertical = i % 2 == 0
                horizontal = j % 2 == 0
                if vertical:
                    if horizontal:
                        result.append("+")
                    else:
                        result.append("|")
                elif horizontal:
                    result.append("-")
                else:
                    result.append(" ")
            result.append("\n")

        labels: Dict[Tuple[Tuple[Perm, ...], bool], str] = {}

        # Put the sets in the tiles

        # How many characters are in a row in the grid
        row_width = 2 * dim_i + 2
        curr_label = 1
        for cell, gridded_perms in sorted(self.cell_basis().items()):
            obstructions, _ = gridded_perms
            basis = list(sorted(obstructions))
            if basis == [Perm((0,))]:
                continue
            # the block, is the basis and whether or not positive
            block = (tuple(basis), cell in self.positive_cells)
            label = labels.get(block)
            if label is None:
                if basis == [Perm((0, 1)), Perm((1, 0))]:
                    if cell in self.positive_cells:
                        label = "\u25cf"
                    else:
                        label = "\u25cb"
                elif basis == [Perm((0, 1))]:
                    label = "\\"
                elif basis == [Perm((1, 0))]:
                    label = "/"
                else:
                    label = str(curr_label)
                    curr_label += 1
                labels[block] = label
            row_index_from_top = dim_j - cell[1] - 1
            index = (2 * row_index_from_top + 1) * row_width + 2 * cell[0] + 1
            result[index] = label

        # Legend at bottom
        for block, label in sorted(labels.items(), key=lambda x: x[1]):
            basis_el, positive = block
            result.append(label)
            result.append(": ")
            if basis_el == (Perm((0, 1)), Perm((1, 0))) and positive:
                result.append("point")
            else:
                result.append(
                    f"Av{'+' if positive else ''}"
                    f"({', '.join(str(p) for p in basis_el)})"
                )
            result.append("\n")

        if any(not ob.is_single_cell() for ob in self.obstructions):
            result.append("Crossing obstructions:\n")
            for ob in self.obstructions:
                if not ob.is_single_cell():
                    result.append(str(ob))
                    result.append("\n")
        for i, req in enumerate(self.requirements):
            result.append(f"Requirement {i}:\n")
            for r in req:
                result.append(str(r))
                result.append("\n")
        for i, ass in enumerate(self.assumptions):
            result.append(f"Assumption {i}:\n")
            result.append(str(ass))
            result.append("\n")
        if self.assumptions or self.requirements:
            result = result[:-1]

        return "".join(result)



from collections import defaultdict
from itertools import chain, islice
from random import randint
from typing import Callable, Dict, Iterator, List, Optional, Set, Tuple, cast

from comb_spec_searcher import Constructor, Strategy, StrategyFactory
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies import NonBijectiveRule, Rule
from comb_spec_searcher.typing import Objects
from tilings import GriddedPerm, Tiling
from tilings.algorithms import Fusion

from ..pointing import DivideByK
from .constructor import FusionConstructor, ReverseFusionConstructor


class FusionRule(NonBijectiveRule[Tiling, GriddedPerm]):
    """Overwritten the generate objects of size method, as this relies on
    knowing the number of left and right points of the parent tiling."""

    @property
    def strategy(self) -> "FusionStrategy":
        return cast(
            FusionStrategy,
            super().strategy,
        )

    @property
    def constructor(self) -> FusionConstructor:
        return cast(FusionConstructor, super().constructor)

    def is_equivalence(
        self, is_empty: Optional[Callable[[Tiling], bool]] = None
    ) -> bool:
        return False

    def _ensure_level_objects(self, n: int) -> None:

        if self.subobjects is None:
            raise RuntimeError("set_subrecs must be set first")
        while n >= len(self.objects_cache):
            res: Objects = defaultdict(list)
            min_left, min_right = self.constructor.min_points

            def add_new_gp(
                params: List[int],
                left_points: int,
                fuse_region_points: int,
                unfused_gps: Iterator[GriddedPerm],
            ) -> None:
                """Update new terms if there is enough points on the left and right."""
                gp = next(unfused_gps)
                if (
                    min_left <= left_points
                    and min_right <= fuse_region_points - left_points
                ):
                    res[tuple(params)].append(gp)

            for param, objects in self.subobjects[0](len(self.objects_cache)).items():
                fuse_region_points = param[self.constructor.fuse_parameter_index]
                for gp in objects:
                    new_params = list(self.constructor.children_param_map(param))
                    unfused_gps = self.strategy.backward_map(
                        self.comb_class, (gp,), self.children
                    )  # iterates over unfused gridded perms in order
                    # with 0, 1, .., and finally fuse_region_points on the left
                    for idx in self.constructor.left_parameter_indices:
                        new_params[idx] -= fuse_region_points
                    add_new_gp(new_params, 0, fuse_region_points, unfused_gps)
                    for left_points in range(1, fuse_region_points + 1):
                        for idx in self.constructor.left_parameter_indices:
                            new_params[idx] += 1
                        for idx in self.constructor.right_parameter_indices:
                            new_params[idx] -= 1
                        add_new_gp(
                            new_params, left_points, fuse_region_points, unfused_gps
                        )

            self.objects_cache.append(res)

    def random_sample_object_of_size(self, n: int, **parameters: int) -> GriddedPerm:
        """Return a random objects of the give size."""
        assert (
            self.subrecs is not None and self.subsamplers is not None
        ), "you must call the set_subrecs function first"
        subrec = self.subrecs[0]
        subsampler = self.subsamplers[0]
        parent_count = self.count_objects_of_size(n, **parameters)
        random_choice = randint(1, parent_count)
        total = 0
        left_right_points = self.constructor.determine_number_of_points_in_fuse_region(
            n, **parameters
        )
        for left_points, right_points in left_right_points:
            new_params = self.constructor.update_subparams(
                left_points, right_points, **parameters
            )
            if new_params is not None:
                assert (
                    new_params[self.constructor.fuse_parameter]
                    == left_points + right_points
                )
                total += subrec(n, **new_params)
                if random_choice <= total:
                    gp = subsampler(n, **new_params)
                    try:
                        return next(
                            self.strategy.backward_map(
                                self.comb_class, (gp,), self.children, left_points
                            )
                        )
                    except StopIteration:
                        assert 0, "something went wrong"
        raise RuntimeError("The for-loop for randomly sampling objects was empty")

    def _forward_order(
        self,
        obj: GriddedPerm,
        image: Tuple[Optional[GriddedPerm], ...],
        data: Optional[object] = None,
    ) -> int:
        # The position of the original object in the backward map of the child object
        return next(i for i, gp in enumerate(self.backward_map(image)) if gp == obj)

    def _backward_order_item(
        self,
        idx: int,
        objs: Tuple[Optional[GriddedPerm], ...],
        data: Optional[object] = None,
    ) -> GriddedPerm:
        if data:  # reverse order
            return tuple(self.backward_map(objs))[-idx - 1]
        return next(islice(self.backward_map(objs), idx, None))


class FusionStrategy(Strategy[Tiling, GriddedPerm]):
    def __init__(self, row_idx=None, col_idx=None, tracked: bool = False):
        self.col_idx = col_idx
        self.row_idx = row_idx
        self.tracked = tracked
        if not sum(1 for x in (self.col_idx, self.row_idx) if x is not None) == 1:
            raise RuntimeError("Cannot specify a row and a column")
        super().__init__(
            ignore_parent=False, inferrable=True, possibly_empty=False, workable=True
        )

    def __call__(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> FusionRule:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        return FusionRule(self, comb_class, children=children)

    def fusion_algorithm(self, tiling: Tiling) -> Fusion:
        return Fusion(
            tiling, row_idx=self.row_idx, col_idx=self.col_idx, tracked=self.tracked
        )

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        algo = self.fusion_algorithm(comb_class)
        if algo.fusable():
            return (algo.fused_tiling(),)
        raise AttributeError("Trying to fuse a tiling that does not fuse")

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling):
        return False

    def is_reversible(self, comb_class: Tiling) -> bool:
        algo = self.fusion_algorithm(comb_class)
        new_ass = algo.new_assumption()
        fused_assumptions = (
            ass.__class__(gps)
            for ass, gps in zip(comb_class.assumptions, algo.assumptions_fuse_counters)
        )
        return new_ass in fused_assumptions

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[int, ...]:
        return (0,)

    def constructor(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> FusionConstructor:
        if not self.tracked:
            # constructor only enumerates when tracked.
            raise NotImplementedError("The fusion strategy was not tracked.")
        # Need to recompute some info to count, so ignoring passed in children
        algo = self.fusion_algorithm(comb_class)
        if not algo.fusable():
            raise StrategyDoesNotApply("Strategy does not apply")
        child = algo.fused_tiling()
        assert children is None or children == (child,)
        min_left, min_right = algo.min_left_right_points()
        return FusionConstructor(
            comb_class,
            child,
            self._fuse_parameter(comb_class),
            self.extra_parameters(comb_class, children)[0],
            *self.left_right_both_sided_parameters(comb_class),
            min_left,
            min_right,
        )

    def reverse_constructor(  # pylint: disable=no-self-use
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        if not self.tracked:
            # constructor only enumerates when tracked.
            raise NotImplementedError("The fusion strategy was not tracked.")
        if children is None:
            children = self.decomposition_function(comb_class)
        # Need to recompute some info to count, so ignoring passed in children
        algo = self.fusion_algorithm(comb_class)
        if not algo.fusable():
            raise StrategyDoesNotApply("Strategy does not apply")
        child = algo.fused_tiling()
        assert children is None or children == (child,)
        (
            left_sided_params,
            right_sided_params,
            _,
        ) = self.left_right_both_sided_parameters(comb_class)
        if not left_sided_params and not right_sided_params:
            if algo.min_left_right_points() != (0, 0):
                raise NotImplementedError(
                    "Reverse positive fusion counting not implemented"
                )
            fused_assumption = algo.new_assumption()
            unfused_assumption = fused_assumption.__class__(
                chain.from_iterable(
                    algo.unfuse_gridded_perm(gp) for gp in fused_assumption.gps
                )
            )
            assert unfused_assumption in comb_class.assumptions
            return DivideByK(
                comb_class,
                children,
                1,
                comb_class.get_assumption_parameter(unfused_assumption),
                self.extra_parameters(comb_class, children),
            )
        left_points, right_points = algo.min_left_right_points()
        return ReverseFusionConstructor(
            comb_class,
            child,
            self._fuse_parameter(comb_class),
            self.extra_parameters(comb_class, children)[0],
            tuple(left_sided_params),
            tuple(right_sided_params),
            left_points,
            right_points,
        )

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str]]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        algo = self.fusion_algorithm(comb_class)
        child = children[0]
        mapped_assumptions = [
            child.forward_map.map_assumption(ass.__class__(gps))
            for ass, gps in zip(comb_class.assumptions, algo.assumptions_fuse_counters)
        ]
        return (
            {
                k: child.get_assumption_parameter(ass)
                for k, ass in zip(comb_class.extra_parameters, mapped_assumptions)
                if ass.gps
            },
        )

    def left_right_both_sided_parameters(
        self, comb_class: Tiling
    ) -> Tuple[Set[str], Set[str], Set[str]]:
        left_sided_params: Set[str] = set()
        right_sided_params: Set[str] = set()
        both_sided_params: Set[str] = set()
        algo = self.fusion_algorithm(comb_class)
        for assumption in comb_class.assumptions:
            parent_var = comb_class.get_assumption_parameter(assumption)
            left_sided = algo.is_left_sided_assumption(assumption)
            right_sided = algo.is_right_sided_assumption(assumption)
            if left_sided and not right_sided:
                left_sided_params.add(parent_var)
            elif right_sided and not left_sided:
                right_sided_params.add(parent_var)
            elif not left_sided and not right_sided:
                both_sided_params.add(parent_var)
        return (
            left_sided_params,
            right_sided_params,
            both_sided_params,
        )

    def _fuse_parameter(self, comb_class: Tiling) -> str:
        algo = self.fusion_algorithm(comb_class)
        child = algo.fused_tiling()
        ass = algo.new_assumption()
        fuse_assumption = ass.__class__(child.forward_map.map_gp(gp) for gp in ass.gps)
        return child.get_assumption_parameter(fuse_assumption)

    def formal_step(self) -> str:
        fusing = "rows" if self.row_idx is not None else "columns"
        idx = self.row_idx if self.row_idx is not None else self.col_idx
        return f"fuse {fusing} {idx} and {idx+1}"

    # pylint: disable=arguments-differ
    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
        left_points: Optional[int] = None,
    ) -> Iterator[GriddedPerm]:
        """
        The backward direction of the underlying bijection used for object
        generation and sampling.
        """
        if children is None:
            children = self.decomposition_function(comb_class)
        gp = objs[0]
        assert gp is not None
        gp = children[0].backward_map.map_gp(gp)
        yield from self.fusion_algorithm(comb_class).unfuse_gridded_perm(
            gp, left_points
        )

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        """
        The forward direction of the underlying bijection used for object
        generation and sampling.
        """
        if children is None:
            children = self.decomposition_function(comb_class)
        fused_gp = self.fusion_algorithm(comb_class).fuse_gridded_perm(obj)
        return (children[0].forward_map.map_gp(fused_gp),)

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("ignore_parent")
        d.pop("inferrable")
        d.pop("possibly_empty")
        d.pop("workable")
        d["row_idx"] = self.row_idx
        d["col_idx"] = self.col_idx
        d["tracked"] = self.tracked
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "FusionStrategy":
        return cls(**d)

    @staticmethod
    def get_op_symbol() -> str:
        return "⚮"

    @staticmethod
    def get_eq_symbol() -> str:
        return "↣"

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + f"(row_idx={self.row_idx}, col_idx={self.col_idx}, "
            f"tracked={self.tracked})"
        )


class FusionFactory(StrategyFactory[Tiling]):
    def __init__(self, tracked: bool = True, isolation_level: Optional[str] = None):
        self.tracked = tracked
        self.isolation_level = isolation_level

    def make_tracked(self) -> "FusionFactory":
        """Return a new version of the strategy that is tracked."""
        if self.tracked:
            raise ValueError("FusionFactory already tracked")
        return self.__class__(tracked=True, isolation_level=self.isolation_level)

    def __call__(self, comb_class: Tiling) -> Iterator[Rule]:
        cols, rows = comb_class.dimensions
        for row_idx in range(rows - 1):
            algo = Fusion(
                comb_class,
                row_idx=row_idx,
                tracked=self.tracked,
                isolation_level=self.isolation_level,
            )
            if algo.fusable():
                fused_tiling = algo.fused_tiling()
                yield FusionStrategy(row_idx=row_idx, tracked=self.tracked)(
                    comb_class, (fused_tiling,)
                )
        for col_idx in range(cols - 1):
            algo = Fusion(
                comb_class,
                col_idx=col_idx,
                tracked=self.tracked,
                isolation_level=self.isolation_level,
            )
            if algo.fusable():
                fused_tiling = algo.fused_tiling()
                yield FusionStrategy(col_idx=col_idx, tracked=self.tracked)(
                    comb_class, (fused_tiling,)
                )

    def __str__(self) -> str:
        if self.tracked:
            return "tracked fusion"
        return "fusion"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(tracked={self.tracked})"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["tracked"] = self.tracked
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "FusionFactory":
        return cls(**d)



import abc
from collections import deque
from itertools import chain
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional

import requests
from sympy import Expr, Symbol, diff, simplify, sympify, var

from comb_spec_searcher.utils import taylor_expand
from permuta import Av
from permuta.permutils.symmetry import lex_min
from tilings.exception import InvalidOperationError
from tilings.griddedperm import GriddedPerm
from tilings.misc import is_tree

if TYPE_CHECKING:
    from tilings import Tiling

x = Symbol("x")


class Enumeration(abc.ABC):
    """
    General representation of a strategy to enumerate tilings.
    """

    def __init__(self, tiling: "Tiling"):
        self.tiling = tiling

    @abc.abstractmethod
    def verified(self) -> bool:
        """
        Returns True if enumeration strategy works for the tiling.
        """
        raise NotImplementedError

    def get_genf(self, **kwargs) -> Expr:
        """
        Returns the generating function for the tiling.

        Raises an InvalidOperationError if the tiling is not verified.
        """
        if not self.verified():
            raise InvalidOperationError("The tiling is not verified")
        raise NotImplementedError

    def __repr__(self) -> str:
        return "Enumeration for:\n" + str(self.tiling)


class LocalEnumeration(Enumeration):
    """
    Enumeration strategy for a locally enumerable tiling.

    A tiling is locally enumerable if the tiling has no crossing obstructions
    or requirements.

    There's not universal way of describing a tiling that is locally enumerable
    with a specification.
    """

    def __init__(self, tiling, no_req=False):
        super().__init__(tiling)
        self.no_req = no_req

    def verified(self) -> bool:
        if self.no_req and self.tiling.requirements:
            return False
        return (
            all(gp.is_single_cell() for gp in self.tiling.obstructions)
            and all(self._req_is_single_cell(req) for req in self.tiling.requirements)
            and all(
                gp.is_single_cell()
                for gp in chain.from_iterable(
                    ass.gps for ass in self.tiling.assumptions
                )
            )
        )

    @staticmethod
    def _req_is_single_cell(req: Iterable[GriddedPerm]) -> bool:
        """
        Returns True if all the gridded perm in the iterable are single cell and in
        the same cell.
        """
        req_iter = iter(req)
        gp0 = next(req_iter)
        if not gp0.is_single_cell():
            return False
        cell = gp0.pos[0]
        all_cells = chain.from_iterable(gp.pos for gp in req_iter)
        return all(c == cell for c in all_cells)

    def get_genf(self, **kwargs) -> Any:
        # pylint: disable=too-many-return-statements
        if not self.verified():
            raise InvalidOperationError("The tiling is not verified")

        funcs: Optional[Dict["Tiling", Any]] = kwargs.get("funcs")
        if funcs is None:
            funcs = {}
        if self.tiling.requirements:
            reqs = self.tiling.requirements[0]
            avoided = self.tiling.__class__(
                self.tiling.obstructions + reqs,
                self.tiling.requirements[1:],
                self.tiling.assumptions,
            )
            without = self.tiling.__class__(
                self.tiling.obstructions,
                self.tiling.requirements[1:],
                self.tiling.assumptions,
            )
            avgf = LocalEnumeration(avoided).get_genf(funcs=funcs)
            wogf = LocalEnumeration(without).get_genf(funcs=funcs)
            return wogf - avgf
        if self.tiling in funcs:
            return funcs[self.tiling]
        # also return something entirely different if the root class/not verified
        if self.tiling.dimensions == (1, 1):
            if self.tiling.is_epsilon():
                return 1
            if self.tiling == self.tiling.__class__.from_string("01_10"):
                return 1 + x
            basis = [ob.patt for ob in self.tiling.obstructions]
            basis_str = "_".join(map(str, lex_min(basis)))
            uri = f"https://permpal.com/perms/raw_data_json/basis/{basis_str}"
            request = requests.get(uri, timeout=10)
            if request.status_code == 404:
                raise NotImplementedError(f"No entry on permpal for {Av(basis)}")
            data = request.json()
            if data["generating_function_sympy"] is None:
                raise NotImplementedError(
                    f"No explicit generating function on permpal for {Av(basis)}"
                )
            return sympify(data["generating_function_sympy"])
        gf = None
        if MonotoneTreeEnumeration(self.tiling).verified():
            gf = MonotoneTreeEnumeration(self.tiling).get_genf()
        if gf is not None:
            funcs[self.tiling] = gf
            return gf
        # TODO: should this create a spec as in the strategy?
        raise NotImplementedError(
            f"Not sure how to enumerate the tiling:\n{self.tiling}"
        )


class MonotoneTreeEnumeration(Enumeration):
    """
    Enumeration strategy for a monotone tree tiling.

    A tiling is a monotone tree if it is local, its cell graph is a tree and
    all but possibly one cell are monotone.

    A monotone tree tiling can be described by a tree where the verified object
    are the cells of the tiling.
    """

    _tracking_var = var("t")

    def verified(self):
        no_req_list = all(len(rl) == 1 for rl in self.tiling.requirements)
        num_non_monotone = sum(
            1 for c in self.tiling.active_cells if not self.tiling.is_monotone_cell(c)
        )
        return (
            self.tiling.dimensions != (1, 1)
            and LocalEnumeration(self.tiling).verified()
            and no_req_list
            and num_non_monotone <= 1
            and is_tree(self.tiling.active_cells, self.tiling.cell_graph())
        )

    def _cell_tree_traversal(self, start):
        """
        Traverse the tree by starting at `start` and always visiting an entire
        row or column before going somewhere else.

        The start vertices is not yielded.
        """
        queue = deque(
            chain(
                self.tiling.cells_in_col(start[0]), self.tiling.cells_in_row(start[1])
            )
        )
        visited = set([start])
        while queue:
            cell = queue.popleft()
            if cell not in visited:
                yield cell
                visited.add(cell)
                queue.extend(self.tiling.cells_in_row(cell[1]))
                queue.extend(self.tiling.cells_in_col(cell[0]))

    def _visted_cells_aligned(self, cell, visited):
        """
        Return the cells that are in visited and in the same row or column as
        `cell`.
        """
        row_cells = self.tiling.cells_in_row(cell[1])
        col_cells = self.tiling.cells_in_col(cell[0])
        return (c for c in visited if (c in row_cells or c in col_cells))

    def get_genf(self, **kwargs) -> Any:
        # pylint: disable=too-many-locals
        if not self.verified():
            raise InvalidOperationError("The tiling is not verified")
        if self.tiling.extra_parameters:
            raise NotImplementedError(
                "Not implemented monotone verified with extra parameters."
            )
        try:
            start = next(
                c
                for c in self.tiling.active_cells
                if not self.tiling.is_monotone_cell(c)
            )
        except StopIteration:
            start = next(iter(self.tiling.active_cells))
        start_basis = self.tiling.cell_basis()[start][0]
        start_reqs = [[p] for p in self.tiling.cell_basis()[start][1]]
        start_tiling = self.tiling.from_perms(
            obstructions=start_basis, requirements=start_reqs
        )
        start_gf = start_tiling.get_genf()
        F = start_gf.subs({x: x * self._cell_variable(start)})
        visited = set([start])
        for cell in self._cell_tree_traversal(start):
            interleaving_cells = self._visted_cells_aligned(cell, visited)
            substitutions = {
                scv: scv * self._tracking_var
                for scv in map(self._cell_variable, interleaving_cells)
            }
            F_tracked = F.subs(substitutions)
            minlen, maxlen = self._cell_num_point(cell)
            if maxlen is None:
                F = self._interleave_any_length(F_tracked, cell)
                if minlen > 0:
                    F -= self._interleave_fixed_lengths(F_tracked, cell, 0, minlen - 1)
            else:
                F = self._interleave_fixed_lengths(F_tracked, cell, minlen, maxlen)
            visited.add(cell)
        F = simplify(
            F.subs({v: 1 for v in F.free_symbols if v != x})
        )  # type: ignore[operator]
        # A simple test to warn us if the code is wrong
        if __debug__:
            lhs = taylor_expand(F, n=6)
            rhs = [len(list(self.tiling.objects_of_size(i))) for i in range(7)]
            assert lhs == rhs, f"Bad genf\n{lhs}\n{rhs}"
        return F

    @staticmethod
    def _cell_variable(cell):
        """
        Return the appropriate variable to track the number of point in the
        given cell.
        """
        return var(f"y_{cell[0]}_{cell[1]}")

    def _interleave_any_length(self, F, cell):
        """
        Return the generating function for interleaving any number of point of
        a monotone sequence into the region tracked by
        `MonotoneTreeEnumeration._tracking_var` in `F`.
        A variable is added to track the number of point in cell.
        """
        cell_var = self._cell_variable(cell)
        gap_filler = 1 / (1 - x * cell_var)
        return F.subs({self._tracking_var: gap_filler}) * gap_filler

    def _interleave_fixed_lengths(self, F, cell, min_length, max_length):
        """
        Return the generating function for interleaving between min_point and
        max_point (both included) number of point of
        a monotone sequence into the region tracked by
        `MonotoneTreeEnumeration._tracking_var` in `F`.
        A variable is added to track the number of point in cell.
        """
        return sum(
            self._interleave_fixed_length(F, cell, i)
            for i in range(min_length, max_length + 1)
        )

    def _interleave_fixed_length(self, F, cell, num_point):
        """
        Return the generating function for interleaving num_point
        number of point of a monotone sequence into the region tracked by
        `MonotoneTreeEnumeration._tracking_var` in `F`.
        A variable is added to track the number of point in cell.
        """
        new_genf = self._tracking_var**num_point * F
        for i in range(1, num_point + 1):
            new_genf = diff(new_genf, self._tracking_var) / i
        new_genf *= self._cell_variable(cell) ** num_point
        new_genf *= x**num_point
        return new_genf.subs({self._tracking_var: 1})

    def _cell_num_point(self, cell):
        """
        Return a pair of integer `(min, max)` that describe the possible
        number of point in the cell. If the number of point is unbounded,
        `max` is None.

        We assume that the cell is monotone
        """
        obs, reqs = self.tiling.cell_basis()[cell]
        ob_lens = sorted(map(len, obs))
        assert ob_lens[0] == 2, "Unexpected obstruction"
        assert len(reqs) <= 1, "Unexpected number of requirement"
        if len(obs) == 1:
            maxlen = None
        elif len(obs) == 2:
            maxlen = ob_lens[1] - 1
        else:
            raise RuntimeError("Unexpected number of obstructions")
        if not reqs:
            minlen = 0
        elif len(reqs) == 1:
            minlen = len(reqs[0])
        else:
            raise RuntimeError("Unexpected number of requirements")
        return minlen, maxlen



from collections import defaultdict, deque
from itertools import chain, product
from typing import (
    TYPE_CHECKING,
    Callable,
    DefaultDict,
    Deque,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
)

from tilings import GriddedPerm, TrackingAssumption

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from tilings import Tiling

COL_INFO = DefaultDict[int, DefaultDict[int, Dict[GriddedPerm, List[int]]]]


class Sliding:
    def __init__(self, tiling: "Tiling") -> None:
        self.tiling = tiling
        self.col_info: COL_INFO = Sliding._get_col_info(self.tiling)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.tiling == other.tiling

    def slidable_pairs(self) -> Iterable[Tuple[int, int]]:
        """Yield the column pairs possible to slide."""
        for av_12, av_123 in product(*self._fast_filter()):
            if self._slide_check_for_pair(av_12, av_123):
                yield av_12, av_123

    def slide_column(self, av_12: int, av_123: int) -> "Tiling":
        """Slide the columns av_12 and av_123 and return the resulting tiling. This
        assumes all checks have been performed and we are allowed to slide these
        columns.
        """
        n = len(next(iter(self.col_info[av_123][0])))
        to_change = set(
            chain(
                self.col_info[av_12][0],
                self.col_info[av_12][1],
                self.col_info[av_123][0],
                self.col_info[av_123][1],
                self.col_info[av_123][2],
                self.col_info[av_123][n - 1],
            )
        )
        obstructions = {obs for obs in self.tiling.obstructions if obs not in to_change}
        obstructions.update(self._slide_obstructions(n, to_change, av_12, av_123))

        return type(self.tiling)(
            requirements=self.tiling.requirements,
            obstructions=obstructions,
            assumptions=self._swap_assumptions(av_12, av_123),
            derive_empty=False,
            remove_empty_rows_and_cols=False,
            simplify=False,
        )

    @staticmethod
    def slide_assumption(
        assumption: TrackingAssumption,
        c1: int,
        c2: int,
        g_map: Optional[Callable[[GriddedPerm], GriddedPerm]] = None,
        g_inv: Optional[Callable[[GriddedPerm], GriddedPerm]] = None,
    ):
        """Swap (c1,c2) if column is either one."""
        if g_map is None or g_inv is None:
            return TrackingAssumption(
                (
                    GriddedPerm(
                        gp.patt,
                        (
                            (
                                c2
                                if gp.pos[0][0] == c1
                                else (c1 if gp.pos[0][0] == c2 else gp.pos[0][0]),
                                0,
                            ),
                        ),
                    )
                    for gp in assumption.gps
                )
            )
        return TrackingAssumption(
            (
                g_inv(
                    GriddedPerm(
                        gp.patt,
                        (
                            (
                                c2
                                if gp.pos[0][0] == c1
                                else (c1 if gp.pos[0][0] == c2 else gp.pos[0][0]),
                                0,
                            ),
                        ),
                    )
                )
                for gp in (g_map(_gp) for _gp in assumption.gps)
            )
        )

    @staticmethod
    def slide_gp(gp: GriddedPerm, c1: int = 0, c2: int = -1) -> "GriddedPerm":
        """Sliding for gridded permutations within a boundary of 1xn."""
        if c2 < 0:
            c2 = c1 + 1
        perms, cols = Sliding._gp_slide_split(gp, c1, c2)

        if not perms[3]:
            return type(gp)(
                chain(perms[0], perms[2], perms[1], perms[5]),
                (
                    (x, 0)
                    for x in chain(cols[0], cols[1], (c2,) * len(perms[1]), cols[2])
                ),
            )

        while perms[1]:
            val = perms[1].pop()
            r_val, idx = min(
                ((r_val, i) for i, r_val in enumerate(perms[3]) if r_val > val),
                default=(-1, -1),
            )
            if idx != -1:
                perms[3][idx] = val
                perms[4].appendleft(r_val)
            else:
                perms[3].appendleft(val)
                perms[4].appendleft(perms[3].pop())

        return type(gp)(
            chain(perms[0], perms[3], perms[2], perms[4], perms[5]),
            (
                (x, 0)
                for x in chain(
                    cols[0],
                    (c1,) * len(perms[3]),
                    cols[1],
                    (c2,) * len(perms[4]),
                    cols[2],
                )
            ),
        )

    @staticmethod
    def slide_gp_inverse(gp: GriddedPerm, c1: int = 0, c2: int = -1) -> "GriddedPerm":
        """Inverse of sliding for gridded permutations within a boundary of 1xn."""
        if c2 < 0:
            c2 = c1 + 1
        perms, cols = Sliding._gp_slide_split(gp, c1, c2)

        if not perms[1]:
            return type(gp)(
                chain(perms[0], perms[3], perms[2], perms[5]),
                (
                    (x, 0)
                    for x in chain(cols[0], (c1,) * len(perms[3]), cols[1], cols[2])
                ),
            )

        while perms[3]:
            val = perms[3].popleft()
            l_val, idx = max(
                ((l_val, i) for i, l_val in enumerate(perms[1]) if l_val < val),
                default=(-1, -1),
            )
            if idx != -1:
                perms[1][idx] = val
                perms[4].append(l_val)
            else:
                perms[1].append(val)
                perms[4].append(perms[1].popleft())

        return type(gp)(
            chain(perms[0], perms[4], perms[2], perms[1], perms[5]),
            (
                (x, 0)
                for x in chain(
                    cols[0],
                    (c1,) * len(perms[4]),
                    cols[1],
                    (c2,) * len(perms[1]),
                    cols[2],
                )
            ),
        )

    @staticmethod
    def _get_col_info(tiling: "Tiling") -> COL_INFO:
        """Gather data about columns needed for sliding."""
        col_info: COL_INFO = defaultdict(lambda: defaultdict(dict))
        for obstruction in tiling.obstructions:
            lst, indices = obstruction.pos[0][0], []
            for i, (_, (x, _)) in enumerate(obstruction):
                if x == lst:
                    indices.append(i)
                else:
                    col_info[lst][len(indices)][obstruction] = indices
                    indices = [i]
                lst = x
            cnt = 0 if len(indices) == len(obstruction) else len(indices)
            col_info[lst][cnt][obstruction] = indices
        return col_info

    def _fast_filter(self) -> Tuple[Set[int], Set[int]]:
        """Fast initial filter to eliminate as much as possible quickly."""
        c, r = self.tiling.dimensions
        if r > 1:
            return set(), set()
        av_12: Set[int] = set()
        av_123: Set[int] = set()
        self._fast_check_all_columns(c, av_12, av_123)
        reqs = {x for req in chain(*self.tiling.requirements) for x, _ in req.pos}
        return av_12 - reqs, av_123 - reqs

    def _fast_check_all_columns(self, c: int, av_12: Set[int], av_123: Set[int]):
        for col in range(c):
            col_perms = self.col_info[col]
            # All gp with points in column have the same number of points in column
            if len(col_perms) < 2:
                continue
            local = col_perms[0]
            # If there are no or more than one local obstructions
            if len(local) != 1:
                continue
            loc_patt = next(iter(local))
            # If the local obstructions is Av(0)
            # If the local obstruction is not increasing
            # If there are more than 4 types of gps number of points within column
            if (
                len(loc_patt) == 1
                or not loc_patt.patt.is_increasing()
                or len(col_perms) > 4
            ):
                continue
            cnts = sorted(col_perms.keys())
            # If the 2nd most number of pnts from a gp within the column is more than 2
            # Note that all local patterns are stored at col_perms[0].
            if cnts[-2] > 2:
                continue
            n = len(loc_patt)
            if n == 2:
                # If the size of the local obstruction is 2 and the largest number of
                # points from a crossing obstruction is less than 3, then this can
                # possibly be a av_12 column in sliding but not a av_123.
                if cnts[-1] < 3:
                    av_12.add(col)
                continue
            if cnts[-1] == n - 1:
                largest_crossings = col_perms[cnts[-1]]
                # If the largest number of points from a crossing obsturction is n-1
                # this could potentially be a av_123 column. It is not if
                # * there are multipe crossing gp of length n-1 (exception if n=3)
                # * There are no increasing patterns of length n-1
                if (n != 3 and len(largest_crossings) > 1) or all(
                    not p.patt.is_increasing() for p in largest_crossings
                ):
                    continue
                av_123.add(col)

    def _slide_check_for_pair(self, av_12: int, av_123: int):
        """Check if slide can be applied to columns av_12 and av_123. We except them
        to have been passed to the fast filter before checking here.
        """
        check_list = self._check_av12_av123_connection(av_12, av_123)
        if check_list is None:
            return False

        for gp, indices in self.col_info[av_123][1].items():
            check_list.remove(gp)
            idx = indices[0]
            if gp in self.col_info[av_12][1]:
                # If the gp is also in the av_12 column, then there should be a
                # similar gp with two points in av_123
                if not self._av123_two_points(gp, idx, (av_12, av_123), check_list):
                    return False
            else:
                # If the gp is not in av_12 column, than there should be a similar
                # gp but with one point in the av_12 column
                if not self._av12_one_point(gp, idx, av_12, check_list):
                    return False
        return len(check_list) == 0

    def _check_av12_av123_connection(
        self, av_12: int, av_123: int
    ) -> Optional[Set[GriddedPerm]]:
        """Check if there is a 123...n obstruction connecting the two columns and if so,
        generate the other obstructions that must be in place for them to be slidable.
        """
        n = len(next(iter(self.col_info[av_123][0])))
        pos = ((av_123, 0),) * (n - 1)
        pos = pos + ((av_12, 0),) if av_12 > av_123 else ((av_12, 0),) + pos
        conn = GriddedPerm(range(n), pos)
        if (
            conn not in self.col_info[av_12][1]
            or conn not in self.col_info[av_123][n - 1]
        ):
            return None
        check_list: Set[GriddedPerm] = set(
            chain(
                self.col_info[av_123][1],
                self.col_info[av_123][2],
                self.col_info[av_12][1],
            )
        )
        check_list.remove(conn)
        return check_list

    def _av123_two_points(
        self,
        gp: GriddedPerm,
        idx: int,
        av_12_av_123: Tuple[int, int],
        check_list: Set[GriddedPerm],
    ) -> bool:
        """Check if gp exists but with point that is in av_12 in av_123. Returns true if
        that is the case and removes it from the checklist. Otherwise false is returned.
        """
        idx2 = self.col_info[av_12_av_123[0]][1][gp][0]
        a, b = sorted((idx, idx2))
        if gp.patt[a] + 1 != gp.patt[b]:
            return False
        positions = []
        pattern = []
        for i, (v, (x, _)) in enumerate(gp):
            if i == idx:
                positions.append((x, 0))
                positions.append((x, 0))
                pattern.append(gp.patt[a])
                pattern.append(gp.patt[b])
            elif i != idx2:
                positions.append((x, 0))
                pattern.append(v)
        expected_gp = GriddedPerm(pattern, positions)
        if expected_gp not in self.col_info[av_12_av_123[1]][2]:
            return False
        check_list.remove(expected_gp)
        return True

    def _av12_one_point(
        self, gp: GriddedPerm, idx: int, av_12: int, check_list: Set[GriddedPerm]
    ) -> bool:
        """Check if gp exists but with point that is in av_123 in av_12. Returns true if
        that is the case and removes it from the checklist. Otherwise false is returned.
        """
        k = len(gp)
        val = gp.patt[idx]
        positions = list(gp.pos)
        pattern = list(gp.patt)
        positions[idx] = (av_12, 0)
        # Move val and coordinate to the correct place in a new gp.
        while True:
            if idx != 0 and positions[idx][0] < positions[idx - 1][0]:
                positions[idx] = positions[idx - 1]
                positions[idx - 1] = (av_12, 0)
                pattern[idx] = pattern[idx - 1]
                pattern[idx - 1] = val
                idx -= 1
            elif idx != k - 1 and positions[idx][0] > positions[idx + 1][0]:
                positions[idx] = positions[idx + 1]
                positions[idx + 1] = (av_12, 0)
                pattern[idx] = pattern[idx + 1]
                pattern[idx + 1] = val
                idx += 1
            else:
                break
        expected_gp = GriddedPerm(pattern, positions)
        if expected_gp not in self.col_info[av_12][1]:
            return False
        check_list.remove(expected_gp)
        return True

    def _swap_assumptions(self, c1: int, c2: int) -> Iterable[TrackingAssumption]:
        for assumption in self.tiling.assumptions:
            assert all(len(gp) == 1 for gp in assumption.gps)
            yield Sliding.slide_assumption(assumption, c1, c2)

    def _slide_obstructions(
        self, n: int, to_change: Set[GriddedPerm], av_12: int, av_123: int
    ) -> Iterable[GriddedPerm]:
        """Generate all obstructions in the new
        tiling thatbelong to av_12 or av_123.
        """
        # Local patterns swap places
        yield GriddedPerm(range(n), ((av_12, 0),) * n)
        yield GriddedPerm(range(2), ((av_123, 0), (av_123, 0)))
        for gp in chain(self.col_info[av_12][0], self.col_info[av_123][0]):
            to_change.remove(gp)

        while to_change:
            gp = to_change.pop()
            if gp in self.col_info[av_12][1] and gp in self.col_info[av_123][n - 1]:
                # The 12...n obstrudtion that connects av12 and av123
                yield GriddedPerm(
                    range(n),
                    ((av_12, 0),) * (n - 1) + ((av_123, 0),)
                    if av_12 < av_123
                    else ((av_123, 0),) + ((av_12, 0),) * (n - 1),
                )
            elif gp in self.col_info[av_123][2]:
                # The one with two points in av_123 are altered so that the two
                # occurrences in av_123 now happen in av_12
                indices: List[int] = self.col_info[av_123][2][gp]
                vals = [gp.patt[idx] for idx in indices]
                pre: Tuple[List[int], List[Tuple[int, int]]] = ([], [])
                post: Tuple[List[int], List[Tuple[int, int]]] = ([], [])
                for v, (x, _) in gp:
                    if x == av_123:
                        continue
                    if x < av_12:
                        pre[0].append(v)
                        pre[1].append((x, 0))
                    else:
                        post[0].append(v)
                        post[1].append((x, 0))
                yield GriddedPerm(
                    chain(pre[0], vals, post[0]),
                    chain(pre[1], ((av_12, 0), (av_12, 0)), post[1]),
                )
            else:
                # The other ones remain
                # (one point each, one point av_12, one_point av_123)
                yield gp

    @staticmethod
    def _gp_slide_split(
        gp: GriddedPerm, c1: int, c2: int
    ) -> Tuple[
        Tuple[Deque[int], Deque[int], Deque[int], Deque[int], Deque[int], Deque[int]],
        Tuple[List[int], List[int], List[int]],
    ]:
        assert c1 < c2 and min(c1, c2) >= 0 and all(y == 0 for _, (x, y) in gp)
        # Columns not affected
        cols: Tuple[List[int], List[int], List[int]] = ([], [], [])

        def _to_deque_index(x: int) -> int:
            if x < c1:
                cols[0].append(x)
                return 0
            if x == c1:
                return 1
            if x < c2:
                cols[1].append(x)
                return 2
            if x == c2:
                return 3
            cols[2].append(x)
            return 5

        perms: Tuple[
            Deque[int], Deque[int], Deque[int], Deque[int], Deque[int], Deque[int]
        ] = (
            deque([]),  # Columns before
            deque([]),  # Column c1
            deque([]),  # Columns between
            deque([]),  # Column c2
            deque([]),  # 'New' column
            deque([]),  # Columns after
        )

        for val, (x, _) in gp:
            perms[_to_deque_index(x)].append(val)

        return perms, cols



import logging
from typing import Dict, FrozenSet, Optional

from logzero import logger

from comb_spec_searcher import (
    CombinatorialSpecification,
    CombinatorialSpecificationSearcher,
)
from comb_spec_searcher.strategies.constructor import DisjointUnion
from comb_spec_searcher.strategies.rule import Rule, VerificationRule
from comb_spec_searcher.strategies.strategy import VerificationStrategy
from comb_spec_searcher.strategies.strategy_pack import StrategyPack
from comb_spec_searcher.typing import CSSstrategy
from permuta import Av, Perm
from tilings import GriddedPerm, Tiling
from tilings.strategies.detect_components import CountComponent
from tilings.strategies.factor import FactorStrategy

__all__ = ["shift_from_spec"]


class TmpLoggingLevel:
    def __init__(self, level):
        self.tmp_level = level
        self.curent_level = logger.level

    def __enter__(self):
        logger.setLevel(self.tmp_level)

    def __exit__(self, exc_type, exc_value, traceback):
        logger.setLevel(self.curent_level)


class NoBasisVerification(VerificationStrategy[Tiling, GriddedPerm]):
    def __init__(self, symmetries: FrozenSet[FrozenSet[Perm]]) -> None:
        self.symmetries = symmetries
        super().__init__()

    def formal_step(self) -> str:
        cls = Av(min(self.symmetries))
        if len(self.symmetries) == 1:
            return f"No cell is {cls}"
        return f"No cell is a symmetry of {cls}"

    def verified(self, comb_class: Tiling) -> bool:
        cell_bases = (frozenset(obs) for obs, _ in comb_class.cell_basis().values())
        return not bool(self.symmetries.intersection(cell_bases))

    @classmethod
    def from_dict(cls, d: dict) -> CSSstrategy:
        raise NotImplementedError

    def to_jsonable(self) -> dict:
        raise NotImplementedError


def expanded_spec(
    tiling: Tiling, pack: StrategyPack, symmetries: FrozenSet[FrozenSet[Perm]]
) -> CombinatorialSpecification:
    """
    Return a spec where any tiling that does not have the basis in one cell is
    verified.

    A locally factorable tiling can always result in a spec where the
    verified leaves are one by one if we remove the local verification
    and monotone tree verification strategies and instead use the
    interleaving factors. As we only care about the shift, we can
    tailor our packs to find this.
    """
    # pylint: disable=import-outside-toplevel
    from tilings.strategies.verification import (
        LocalVerificationStrategy,
        MonotoneTreeVerificationStrategy,
    )
    from tilings.tilescope import TileScopePack

    pack = TileScopePack(
        initial_strats=pack.initial_strats,
        inferral_strats=pack.inferral_strats,
        expansion_strats=pack.expansion_strats,
        ver_strats=pack.ver_strats,
        name=pack.name,
    )
    pack = pack.remove_strategy(MonotoneTreeVerificationStrategy()).make_interleaving(
        tracked=False, unions=False
    )
    pack = pack.remove_strategy(LocalVerificationStrategy())
    pack = pack.add_verification(NoBasisVerification(symmetries), apply_first=True)
    with TmpLoggingLevel(logging.WARN):
        css = CombinatorialSpecificationSearcher(tiling, pack)
        spec = css.auto_search()
    return spec


def shift_from_spec(
    tiling: Tiling,
    pack: StrategyPack,
    symmetries: FrozenSet[FrozenSet[Perm]],
) -> Optional[int]:
    assert all(symmetries)
    traverse_cache: Dict[Tiling, Optional[int]] = {}
    spec = expanded_spec(tiling, pack, symmetries)

    def traverse(t: Tiling) -> Optional[int]:
        rule = spec.rules_dict[t]
        if t in traverse_cache:
            return traverse_cache[t]
        res: Optional[int]
        if isinstance(rule.strategy, NoBasisVerification):
            res = None
        elif t.dimensions == (1, 1):
            res = 0
        elif isinstance(rule, VerificationRule):
            raise ValueError(
                "this should be unreachable, looks like JP, HU "
                "and CB misunderstood the code."
            )
        elif isinstance(rule, Rule) and isinstance(rule.strategy, FactorStrategy):
            min_points = [len(next(c.minimal_gridded_perms())) for c in rule.children]
            point_sum = sum(min_points)
            shifts = [point_sum - mpoint for mpoint in min_points]
            children_reliance = [traverse(c) for c in rule.children]
            res = min(
                (r + s for r, s in zip(children_reliance, shifts) if r is not None),
                default=None,
            )
        elif isinstance(rule, Rule) and isinstance(
            rule.constructor, (DisjointUnion, CountComponent)
        ):
            children_reliance = [traverse(c) for c in rule.children]
            res = min((r for r in children_reliance if r is not None), default=None)

        else:
            raise NotImplementedError(rule)
        traverse_cache[t] = res
        return res

    return traverse(tiling)



from collections import Counter, defaultdict
from functools import reduce
from itertools import chain
from operator import mul
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Tuple, cast

import requests
from sympy import Eq, Expr, Function, Symbol, collect, degree, solve, sympify, var

from comb_spec_searcher import (
    AtomStrategy,
    CombinatorialClass,
    CombinatorialSpecification,
    StrategyPack,
    VerificationStrategy,
)
from comb_spec_searcher.exception import InvalidOperationError, StrategyDoesNotApply
from comb_spec_searcher.strategies import VerificationRule
from comb_spec_searcher.typing import Objects, Terms
from permuta import Av, Perm
from permuta.permutils import (
    is_insertion_encodable_maximum,
    is_insertion_encodable_rightmost,
    lex_min,
)
from tilings import GriddedPerm, Tiling
from tilings.algorithms import locally_factorable_shift
from tilings.algorithms.enumeration import LocalEnumeration, MonotoneTreeEnumeration
from tilings.assumptions import ComponentAssumption, TrackingAssumption
from tilings.strategies import (
    DetectComponentsStrategy,
    FactorFactory,
    FactorInsertionFactory,
    RemoveRequirementFactory,
    RequirementCorroborationFactory,
    SymmetriesFactory,
)

from .abstract import BasisAwareVerificationStrategy

x = var("x")

__all__ = [
    "BasicVerificationStrategy",
    "OneByOneVerificationStrategy",
    "LocallyFactorableVerificationStrategy",
    "ElementaryVerificationStrategy",
    "LocalVerificationStrategy",
    "MonotoneTreeVerificationStrategy",
]


TileScopeVerificationStrategy = VerificationStrategy[Tiling, GriddedPerm]


class BasicVerificationStrategy(AtomStrategy):
    """
    TODO: can this be moved to the CSS atom strategy?
    """

    def get_terms(self, comb_class: CombinatorialClass, n: int) -> Terms:
        if not isinstance(comb_class, Tiling):
            raise NotImplementedError
        gp = next(comb_class.minimal_gridded_perms())
        if n == len(gp):
            parameters = tuple(
                assumption.get_value(gp) for assumption in comb_class.assumptions
            )
            return Counter([parameters])
        return Counter()

    def get_objects(self, comb_class: CombinatorialClass, n: int) -> Objects:
        if not isinstance(comb_class, Tiling):
            raise NotImplementedError
        res: Objects = defaultdict(list)
        gp = next(comb_class.minimal_gridded_perms())
        if n == len(gp):
            parameters = tuple(
                assumption.get_value(gp) for assumption in comb_class.assumptions
            )
            res[parameters].append(gp)
        return res

    @staticmethod
    def generate_objects_of_size(
        comb_class: CombinatorialClass, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        """
        Verification strategies must contain a method to generate the objects.
        """
        yield from comb_class.objects_of_size(n, **parameters)

    def random_sample_object_of_size(
        self, comb_class: CombinatorialClass, n: int, **parameters: int
    ) -> GriddedPerm:
        """
        Verification strategies must contain a method to sample the objects.
        """
        key = tuple(y for _, y in sorted(parameters.items()))
        if BasicVerificationStrategy().get_terms(comb_class, n).get(key):
            return cast(GriddedPerm, next(comb_class.objects_of_size(n, **parameters)))
        raise (
            NotImplementedError(
                "Verification strategy did not contain a method to sample the objects"
            )
        )

    def get_genf(
        self,
        comb_class: CombinatorialClass,
        funcs: Optional[Dict[CombinatorialClass, Function]] = None,
    ) -> Any:
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("Can't find generating functon for non-atom.")
        if not isinstance(comb_class, Tiling):
            raise NotImplementedError
        cast(Tiling, comb_class)
        gp = next(comb_class.minimal_gridded_perms())
        expected = {"x": len(gp)}
        for assumption in comb_class.assumptions:
            expected[
                comb_class.get_assumption_parameter(assumption)
            ] = assumption.get_value(gp)
        return reduce(mul, [var(k) ** n for k, n in expected.items()], 1)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class OneByOneVerificationRule(VerificationRule[Tiling, GriddedPerm]):
    def get_equation(
        self,
        get_function: Callable[[Tiling], Function],
        funcs: Optional[Dict[Tiling, Function]] = None,
    ) -> Eq:
        # Find the minimal polynomial for the underlying class
        basis = [ob.patt for ob in self.comb_class.obstructions]
        basis_str = "_".join(map(str, lex_min(basis)))
        uri = f"https://permpal.com/perms/raw_data_json/basis/{basis_str}"
        request = requests.get(uri, timeout=10)
        if request.status_code == 404:
            return super().get_equation(get_function, funcs)
        data = request.json()
        min_poly = data["min_poly_maple"]
        if min_poly is None:
            return Eq(
                get_function(self.comb_class),
                self.tiling_to_symbol_eq(self.comb_class),
            )
        min_poly = min_poly.replace("^", "**").replace("F(x)", "F")
        lhs, _ = min_poly.split("=")
        # We now need to worry about the requirements. The min poly we got is
        # for the class with requirements.
        eq = Eq(self.without_req_genf(self.comb_class), get_function(self.comb_class))
        subs = solve([eq], var("F"), dict=True)[0]
        if self.comb_class.assumptions:
            subs["x"] = var("x") * var("k_0")
        res, _ = sympify(lhs).subs(subs, simultaneous=True).as_numer_denom()
        # Pick the unique factor that contains F
        for factor in res.as_ordered_factors():
            if factor.atoms(Function):
                res = factor
        # currently we have 0 = rhs,
        lhs = get_function(self.comb_class)
        if degree(res, lhs) == 1:
            # solve for rational gf
            rhs = solve([Eq(res, 0)], lhs, dict=True)[0][lhs]
        else:
            # or add F to both sides
            rhs = collect(res + lhs, lhs)
        return Eq(lhs, rhs)

    def tiling_to_symbol_eq(self, tiling: Tiling) -> Any:
        """
        Find the equation for the tiling in terms of F_C's, where C are
        permutation classes.
        """
        if tiling.requirements:
            reqs = tiling.requirements[0]
            avoided = tiling.__class__(
                tiling.obstructions + reqs,
                tiling.requirements[1:],
                tiling.assumptions,
            )
            without = tiling.__class__(
                tiling.obstructions,
                tiling.requirements[1:],
                tiling.assumptions,
            )
            return self.tiling_to_symbol_eq(without) - self.tiling_to_symbol_eq(avoided)
        params = self.comb_class.extra_parameters
        x_var = "x"
        if params:
            assert len(params) == 1
            x_var += "*" + params[0]
        basis = [ob.patt for ob in tiling.obstructions]
        one_based_basis = ",".join(("".join(str(i + 1) for i in b) for b in basis))
        return Symbol(f"F_Av({one_based_basis})({x_var})")

    @property
    def no_req_tiling(self) -> Tiling:
        return self.comb_class.__class__(
            self.comb_class.obstructions, tuple(), self.comb_class.assumptions
        )

    def without_req_genf(self, tiling: Tiling):
        """
        Find the equation for the tiling in terms of F, the generating
        function where the reqs are reomoved from tiling.
        """
        if tiling == self.no_req_tiling:
            return var("F")
        if tiling.requirements:
            reqs = tiling.requirements[0]
            avoided = tiling.__class__(
                tiling.obstructions + reqs,
                tiling.requirements[1:],
                tiling.assumptions,
            )
            without = tiling.__class__(
                tiling.obstructions,
                tiling.requirements[1:],
                tiling.assumptions,
            )
            avgf = self.without_req_genf(avoided)
            wogf = self.without_req_genf(without)
            return wogf - avgf
        return LocalEnumeration(tiling).get_genf()


class OneByOneVerificationStrategy(BasisAwareVerificationStrategy):
    def __init__(
        self,
        basis: Optional[Iterable[Perm]] = None,
        symmetry: bool = False,
        ignore_parent: bool = False,
    ):
        super().__init__(basis, symmetry, ignore_parent)
        self._spec: Dict[Tiling, CombinatorialSpecification] = {}

    @staticmethod
    def _spec_from_permpal(tiling: Tiling) -> CombinatorialSpecification:
        basis = [ob.patt for ob in tiling.obstructions]
        basis_str = "_".join(map(str, lex_min(basis)))
        uri = f"https://permpal.com/perms/raw_data_json/basis/{basis_str}"
        request = requests.get(uri, timeout=10)
        if request.status_code == 404:
            raise InvalidOperationError("Can't find spec for one by one verified rule.")
        data = request.json()
        spec_json = data["specs_and_eqs"][0]["spec_json"]
        spec = cast(
            CombinatorialSpecification, CombinatorialSpecification.from_dict(spec_json)
        )
        actual_class = Tiling(tiling.obstructions)
        if spec.root != actual_class:
            for strategy in SymmetriesFactory()(actual_class):
                rule = strategy(actual_class)
                if rule.children[0] == spec.root:
                    break
            else:
                raise InvalidOperationError("Error fixing sym in 1x1")
            rules = [rule] + list(spec.rules_dict.values())
            spec = CombinatorialSpecification(rule.comb_class, rules)
        assert spec.root == Tiling(tiling.obstructions)
        return spec

    def get_specification(
        self, comb_class: Tiling
    ) -> CombinatorialSpecification[Tiling, GriddedPerm]:
        if comb_class not in self._spec:
            try:
                self._spec[comb_class] = super().get_specification(comb_class)
            except InvalidOperationError as e:
                if len(comb_class.requirements) > 1 or comb_class.dimensions != (1, 1):
                    raise e
                self._spec[comb_class] = self._spec_from_permpal(comb_class)
        return self._spec[comb_class]

    def get_complement_spec(self, tiling: Tiling) -> CombinatorialSpecification:
        assert len(tiling.requirements) == 1
        complement = tiling.remove_requirement(tiling.requirements[0]).add_obstructions(
            tiling.requirements[0]
        )
        return self.get_specification(complement)

    def pack(self, comb_class: Tiling) -> StrategyPack:
        if any(isinstance(ass, ComponentAssumption) for ass in comb_class.assumptions):
            return ComponentVerificationStrategy().pack(comb_class)
        # pylint: disable=import-outside-toplevel
        from tilings.tilescope import TileScopePack

        assert comb_class.dimensions == (1, 1)
        basis, _ = comb_class.cell_basis()[(0, 0)]
        if any(
            any(p.contains(patt) for patt in basis)
            for p in [
                Perm((0, 2, 1)),
                Perm((1, 2, 0)),
                Perm((1, 0, 2)),
                Perm((2, 0, 1)),
            ]
        ):
            # subclass of Av(231) or a symmetry, use point placements!
            return TileScopePack.point_and_row_and_col_placements().add_verification(
                BasicVerificationStrategy(), replace=True
            )
        if is_insertion_encodable_maximum(basis):
            return TileScopePack.regular_insertion_encoding(3)
        if is_insertion_encodable_rightmost(basis):
            return TileScopePack.regular_insertion_encoding(2)
        # if it is the class or positive class
        if not comb_class.requirements or (
            len(comb_class.requirements) == 1
            and len(comb_class.requirements[0]) == 1
            and len(comb_class.requirements[0][0]) <= 2
        ):
            if (
                (Perm((0, 1, 2)) in basis or Perm((2, 1, 0)) in basis)
                and all(len(p) <= 4 for p in basis)
                and len(basis) > 1
            ):
                # is a subclass of Av(123) avoiding patterns of length <= 4
                # experimentally showed that such clsses always terminates
                return TileScopePack.row_and_col_placements().add_basis(basis)
        raise InvalidOperationError(
            "Cannot get a specification for one by one verification for "
            f"subclass Av({basis})"
        )

    def __call__(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> OneByOneVerificationRule:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("The combinatorial class is not verified")
        return OneByOneVerificationRule(self, comb_class, children)

    def verified(self, comb_class: Tiling) -> bool:
        if not comb_class.dimensions == (1, 1):
            return False
        if not self.basis:
            return True
        tiling_class = Av([ob.patt for ob in comb_class.obstructions])
        sym_classes = (Av(sym) for sym in self.symmetries)
        is_strict_subclass = any(
            tiling_class.is_subclass(cls) and cls != tiling_class for cls in sym_classes
        )
        return is_strict_subclass

    def get_genf(
        self, comb_class: Tiling, funcs: Optional[Dict[Tiling, Function]] = None
    ) -> Any:
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("tiling not 1x1 verified")
        if len(comb_class.obstructions) == 1 and comb_class.obstructions[0] in (
            GriddedPerm.single_cell((0, 1, 2), (0, 0)),
            GriddedPerm.single_cell((2, 1, 0), (0, 0)),
        ):
            return LocalEnumeration(comb_class).get_genf(funcs=funcs)
        try:
            return super().get_genf(comb_class, funcs)
        except InvalidOperationError:
            return LocalEnumeration(comb_class).get_genf(funcs=funcs)

    def formal_step(self) -> str:
        return "tiling is a subclass of the original tiling"

    def get_terms(self, comb_class: Tiling, n: int) -> Terms:
        terms = super().get_terms(comb_class.remove_assumptions(), n)
        if (
            comb_class.requirements
            and self.get_specification(comb_class).root != comb_class
        ):
            if len(comb_class.requirements) == 1:
                comp_spec = self.get_complement_spec(comb_class.remove_assumptions())
            else:
                raise NotImplementedError(
                    "Not implemented counting for one by one with two or more reqs"
                )
            comp_terms = comp_spec.get_terms(n)
            terms = Counter({tuple(): terms[tuple()] - comp_terms[tuple()]})
        if comb_class.assumptions:
            assert comb_class.assumptions == (TrackingAssumption.from_cells([(0, 0)]),)
            terms = Counter({(n,): terms[tuple()]})
        return terms

    def get_objects(self, comb_class: Tiling, n: int) -> Objects:
        objects = super().get_objects(comb_class, n)
        if comb_class.requirements:
            if len(comb_class.requirements) == 1:
                comp_spec = self.get_complement_spec(comb_class.remove_assumptions())
            else:
                raise NotImplementedError(
                    "Not implemented objects for one by one with two or more reqs"
                )
            comp_objects = comp_spec.get_objects(n)
            objects = defaultdict(
                list,
                {
                    a: list(set(b).difference(comp_objects[a]))
                    for a, b in objects.items()
                },
            )

        if comb_class.assumptions:
            assert comb_class.assumptions == (TrackingAssumption.from_cells([(0, 0)]),)
            objects = defaultdict(list, {(n,): objects[tuple()]})
        return objects

    def random_sample_object_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> GriddedPerm:
        if comb_class.assumptions:
            assert (
                len(comb_class.assumptions) == 1
                and parameters[
                    comb_class.get_assumption_parameter(comb_class.assumptions[0])
                ]
                == n
            )
        while True:
            # Rejection sampling
            gp = super().random_sample_object_of_size(
                Tiling(comb_class.obstructions), n
            )
            if gp in comb_class:
                return gp

    def __str__(self) -> str:
        if not self.basis:
            return "one by one verification"
        return f"One by one subclass of {Av(self.basis)}"


class ComponentVerificationStrategy(TileScopeVerificationStrategy):
    """Enumeration strategy for verifying 1x1s with component assumptions."""

    def pack(self, comb_class: Tiling) -> StrategyPack:
        raise InvalidOperationError("No pack for removing component assumption")

    def verified(self, comb_class: Tiling) -> bool:
        return comb_class.dimensions == (1, 1) and any(
            isinstance(ass, ComponentAssumption) for ass in comb_class.assumptions
        )

    def decomposition_function(
        self, comb_class: Tiling
    ) -> Optional[Tuple[Tiling, ...]]:
        """
        The rule as the root as children if one of the cell of the tiling is the root.
        """
        if self.verified(comb_class):
            return (comb_class.remove_assumptions(),)
        return None

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[int, ...]:
        return (0,)

    def formal_step(self) -> str:
        return "component verified"

    def get_genf(
        self, comb_class: Tiling, funcs: Optional[Dict[Tiling, Function]] = None
    ) -> Expr:
        raise NotImplementedError(
            "Not implemented method to count objects for component verified"
        )

    def get_terms(self, comb_class: Tiling, n: int) -> Terms:
        raise NotImplementedError(
            "Not implemented method to count objects for component verified"
        )

    def generate_objects_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError(
            "Not implemented method to generate objects for component verified tilings"
        )

    def random_sample_object_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> GriddedPerm:
        raise NotImplementedError(
            "Not implemented random sample for component verified tilings"
        )

    def __str__(self) -> str:
        return "component verification"

    @classmethod
    def from_dict(cls, d: dict) -> "ComponentVerificationStrategy":
        return cls(**d)


class LocallyFactorableVerificationStrategy(BasisAwareVerificationStrategy):
    """
    Verification strategy for a locally factorable tiling.

    A tiling is locally factorable if all its obstructions and requirements are
    locally factorable, i.e. each obstruction or requirement use at most one
    cell on each row and column. To be locally factorable, a tiling
    should not be equivalent to a 1x1 tiling.

    A locally factorable tiling can be describe with a specification with only subset
    verified tiling.
    """

    def pack(self, comb_class: Tiling) -> StrategyPack:
        return StrategyPack(
            name="LocallyFactorable",
            initial_strats=[
                FactorFactory(),
                RequirementCorroborationFactory(),
                DetectComponentsStrategy(),
            ],
            inferral_strats=[],
            expansion_strats=[[FactorInsertionFactory()], [RemoveRequirementFactory()]],
            ver_strats=[
                BasicVerificationStrategy(),
                OneByOneVerificationStrategy(
                    basis=self._basis, symmetry=self._symmetry
                ),
                ComponentVerificationStrategy(),
                InsertionEncodingVerificationStrategy(),
                MonotoneTreeVerificationStrategy(no_factors=True),
                LocalVerificationStrategy(no_factors=True),
            ],
        )

    @staticmethod
    def _pack_for_shift(comb_class: Tiling) -> StrategyPack:
        return StrategyPack(
            name="LocallyFactorable",
            initial_strats=[
                FactorFactory(),
                RequirementCorroborationFactory(),
                DetectComponentsStrategy(),
            ],
            inferral_strats=[],
            expansion_strats=[[FactorInsertionFactory()]],
            ver_strats=[
                BasicVerificationStrategy(),
                OneByOneVerificationStrategy(),
                ComponentVerificationStrategy(),
                InsertionEncodingVerificationStrategy(),
                MonotoneTreeVerificationStrategy(no_factors=True),
                LocalVerificationStrategy(no_factors=True),
            ],
        )

    @staticmethod
    def _locally_factorable_obstructions(tiling: Tiling):
        """
        Check if all the obstructions of the tiling are locally factorable.
        """
        return all(not ob.is_interleaving() for ob in tiling.obstructions)

    @staticmethod
    def _locally_factorable_requirements(tiling: Tiling):
        """
        Check if all the requirements of the tiling are locally factorable.
        """
        reqs = chain.from_iterable(tiling.requirements)
        return all(not r.is_interleaving() for r in reqs)

    def verified(self, comb_class: Tiling):
        return (
            not comb_class.dimensions == (1, 1)
            and self._locally_factorable_obstructions(comb_class)
            and self._locally_factorable_requirements(comb_class)
            and all(
                not isinstance(ass, ComponentAssumption)
                or (
                    len(ass.gps) == 1
                    and comb_class.only_cell_in_row_and_col(list(ass.cells)[0])
                )
                for ass in comb_class.assumptions
            )
        )

    def decomposition_function(
        self, comb_class: Tiling
    ) -> Optional[Tuple[Tiling, ...]]:
        """
        The rule as the root as children if one of the cell of the tiling is the root.
        """
        if self.verified(comb_class):
            if not self.basis:
                return ()
            pack = self._pack_for_shift(comb_class)
            sfs = locally_factorable_shift.shift_from_spec(
                comb_class, pack, self.symmetries
            )
            if sfs is not None:
                return (Tiling.from_perms(self.basis),)
            return ()
        return None

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[int, ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply
        if not children:
            return ()
        pack = self._pack_for_shift(comb_class)
        shift = locally_factorable_shift.shift_from_spec(
            comb_class, pack, self.symmetries
        )
        assert shift is not None
        return (shift,)

    def formal_step(self) -> str:
        return "tiling is locally factorable"

    def __str__(self) -> str:
        return "locally factorable verification"


class ElementaryVerificationStrategy(LocallyFactorableVerificationStrategy):
    """
    Verification strategy for elementary tilings.

    A tiling is elementary if each active cell is on its own row and column.
    To be elementary, a tiling should not be equivalent to a 1x1
    tiling.

    By definition an elementary tiling is locally factorable.

    A elementary tiling can be describe with a specification with only one by one
    verified tiling.
    """

    def verified(self, comb_class: Tiling):
        return (
            comb_class.fully_isolated()
            and not comb_class.dimensions == (1, 1)
            and all(
                len(ass.gps) == 1
                for ass in comb_class.assumptions
                if isinstance(ass, ComponentAssumption)
            )
        )

    def formal_step(self) -> str:
        return "tiling is elementary verified"

    @classmethod
    def from_dict(cls, d: dict) -> "ElementaryVerificationStrategy":
        return cls(**d)

    def __str__(self) -> str:
        return "elementary verification"


class LocalVerificationStrategy(BasisAwareVerificationStrategy):
    """
    The local verified strategy.

    A tiling is local verified if every obstruction and every requirement is
    localized, i.e. in a single cell and the tiling is not 1x1.
    """

    def __init__(
        self,
        basis: Optional[Iterable[Perm]] = None,
        symmetry: bool = False,
        ignore_parent: bool = False,
        no_factors: bool = False,
    ):
        self.no_factors = no_factors
        super().__init__(basis, symmetry, ignore_parent)

    def pack(self, comb_class: Tiling) -> StrategyPack:
        try:
            return InsertionEncodingVerificationStrategy().pack(comb_class)
        except StrategyDoesNotApply:
            pass
        if self.no_factors:
            raise InvalidOperationError(
                f"Cannot get a simpler specification for\n{comb_class}"
            )
        return StrategyPack(
            initial_strats=[FactorFactory(), DetectComponentsStrategy()],
            inferral_strats=[],
            expansion_strats=[],
            ver_strats=[
                BasicVerificationStrategy(),
                OneByOneVerificationStrategy(self.basis, self._symmetry),
                ComponentVerificationStrategy(),
                InsertionEncodingVerificationStrategy(),
                MonotoneTreeVerificationStrategy(no_factors=True),
                LocalVerificationStrategy(self.basis, self._symmetry, no_factors=True),
            ],
            name="factor pack",
        )

    def verified(self, comb_class: Tiling) -> bool:
        return (
            comb_class.dimensions != (1, 1)
            and (not self.no_factors or len(comb_class.find_factors()) == 1)
            and LocalEnumeration(comb_class).verified()
            and all(
                not isinstance(ass, ComponentAssumption)
                or (
                    len(ass.gps) == 1
                    and comb_class.only_cell_in_row_and_col(list(ass.cells)[0])
                )
                for ass in comb_class.assumptions
            )
        )

    def formal_step(self) -> str:
        return "tiling is locally enumerable"

    @classmethod
    def from_dict(cls, d: dict) -> "LocalVerificationStrategy":
        return cls(**d)

    def get_genf(
        self, comb_class: Tiling, funcs: Optional[Dict[Tiling, Function]] = None
    ) -> Any:
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("tiling not locally verified")
        if len(comb_class.obstructions) == 1 and comb_class.obstructions[0] in (
            GriddedPerm.single_cell((0, 1, 2), (0, 0)),
            GriddedPerm.single_cell((2, 1, 0), (0, 0)),
        ):
            return LocalEnumeration(comb_class).get_genf(funcs=funcs)
        try:
            return super().get_genf(comb_class, funcs)
        except InvalidOperationError:
            return LocalEnumeration(comb_class).get_genf(funcs=funcs)

    def get_terms(self, comb_class: Tiling, n: int) -> Terms:
        raise NotImplementedError(
            "Not implemented method to count objects for locally verified tilings"
        )

    def generate_objects_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError(
            "Not implemented method to generate objects for locally verified tilings"
        )

    def random_sample_object_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> GriddedPerm:
        raise NotImplementedError(
            "Not implemented random sample for locally verified tilings"
        )

    def __str__(self) -> str:
        return "local verification"


class InsertionEncodingVerificationStrategy(TileScopeVerificationStrategy):
    """
    Verify all n x 1 and 1 x n tilings that have a regular insertion encoding.
    """

    def __init__(self, ignore_parent: bool = False):
        super().__init__(ignore_parent=ignore_parent)

    def pack(self, comb_class: Tiling) -> StrategyPack:
        # pylint: disable=import-outside-toplevel
        from tilings.strategy_pack import TileScopePack

        if self.has_rightmost_insertion_encoding(comb_class):
            pack = TileScopePack.regular_insertion_encoding(2)
        elif self.has_topmost_insertion_encoding(comb_class):
            pack = TileScopePack.regular_insertion_encoding(3)
        else:
            raise StrategyDoesNotApply(
                "tiling does not has a regular insertion encoding"
            )
        return pack.add_initial(DetectComponentsStrategy())

    @staticmethod
    def has_rightmost_insertion_encoding(tiling: Tiling) -> bool:
        return tiling.dimensions[0] == 1 and all(
            is_insertion_encodable_rightmost(basis)
            for basis, _ in tiling.cell_basis().values()
        )

    @staticmethod
    def has_topmost_insertion_encoding(tiling: Tiling) -> bool:
        return tiling.dimensions[1] == 1 and all(
            is_insertion_encodable_maximum(basis)
            for basis, _ in tiling.cell_basis().values()
        )

    def verified(self, comb_class: Tiling) -> bool:
        return self.has_rightmost_insertion_encoding(
            comb_class
        ) or self.has_topmost_insertion_encoding(comb_class)

    def formal_step(self) -> str:
        return "tiling has a regular insertion encoding"

    @classmethod
    def from_dict(cls, d: dict) -> "InsertionEncodingVerificationStrategy":
        return cls(**d)

    def get_terms(self, comb_class: Tiling, n: int) -> Terms:
        raise NotImplementedError(
            "Not implemented method to count objects for insertion encoding "
            "verified tilings"
        )

    def generate_objects_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError(
            "Not implemented method to generate objects for insertion encoding "
            "verified tilings"
        )

    def random_sample_object_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> GriddedPerm:
        raise NotImplementedError(
            "Not implemented random sample for insertion encoding verified tilings"
        )

    def __str__(self) -> str:
        return "insertion encoding verified"


class MonotoneTreeVerificationStrategy(TileScopeVerificationStrategy):
    """
    Verify all tiling that is a monotone tree.
    """

    def __init__(self, ignore_parent: bool = False, no_factors: bool = True):
        self.no_factors = no_factors
        super().__init__(ignore_parent=ignore_parent)

    def pack(self, comb_class: Tiling) -> StrategyPack:
        try:
            return InsertionEncodingVerificationStrategy().pack(comb_class)
        except StrategyDoesNotApply:
            pass
        if self.no_factors:
            raise InvalidOperationError("Cannot get a simpler specification")
        return StrategyPack(
            initial_strats=[FactorFactory()],
            inferral_strats=[],
            expansion_strats=[],
            ver_strats=[
                BasicVerificationStrategy(),
                OneByOneVerificationStrategy(),
                InsertionEncodingVerificationStrategy(),
                MonotoneTreeVerificationStrategy(no_factors=True),
            ],
            name="factor pack",
        )

    def verified(self, comb_class: Tiling) -> bool:
        return (
            not self.no_factors or len(comb_class.find_factors()) == 1
        ) and MonotoneTreeEnumeration(comb_class).verified()

    def formal_step(self) -> str:
        return "tiling is a monotone tree"

    @classmethod
    def from_dict(cls, d: dict) -> "MonotoneTreeVerificationStrategy":
        return cls(**d)

    def get_genf(
        self, comb_class: Tiling, funcs: Optional[Dict[Tiling, Function]] = None
    ) -> Any:
        if not self.verified(comb_class):
            raise StrategyDoesNotApply("tiling not locally verified")
        try:
            return super().get_genf(comb_class, funcs)
        except InvalidOperationError:
            return MonotoneTreeEnumeration(comb_class).get_genf(funcs=funcs)

    def get_terms(self, comb_class: Tiling, n: int) -> Terms:
        raise NotImplementedError(
            "Not implemented method to count objects for monotone tree "
            "verified tilings"
        )

    def generate_objects_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError(
            "Not implemented method to generate objects for monotone tree "
            "verified tilings"
        )

    def random_sample_object_of_size(
        self, comb_class: Tiling, n: int, **parameters: int
    ) -> GriddedPerm:
        raise NotImplementedError(
            "Not implemented random sample for monotone tree verified tilings"
        )

    def __str__(self) -> str:
        return "monotone tree verification"
